/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/



class Vector{
  REAL data[SIZEVECT+1];
 public:
  int size;
  Vector();
  Vector(int s);
  Vector(Vector *);
  Vector(double d);
  Vector operator +(const Vector &rhs);
  Vector operator *(const Vector &rhs);
  Vector operator -(const Vector &rhs);
  Vector operator -();
  Vector operator /(const Vector &rhs);
  Vector &operator =(const Vector &rhs);
  Vector &operator +=(const Vector &rhs);
  void allocData(int s);
  double Dot(Vector &arg1);
  void setValue(int i, double x);
  double value(int i);
  void load(double x); // sets whole vect to value.
  void print();
};

Vector::Vector(){
  size = SIZEVECT;
};

Vector::Vector(double x){
  int i;
  size = SIZEVECT;
  // printf("(Vector)%d\n", size);
  for (i = 0; i< SIZEVECT; i++)
    data[i] = x;
}

Vector::Vector(int){
   size = SIZEVECT;
}

Vector::Vector(Vector *v){
  int i;
  size = v->size;
  for (i = 0; i< SIZEVECT; i++)
    data[i] = v->data[i];
}

void Vector ::allocData(int s){
  int i;
  size = s;
  for (i = 0; i< SIZEVECT; i++)
    data[i] = 0;
}

REAL Vector::Dot(Vector &arg1)
{
  int i;
  REAL x;
  REAL *p1, *p2;
  x = 0.0;
  p1 = &(this->data[0]);
  p2 = &(arg1.data[0]);
// modified to avoid averflow
  for(i = 0; i < size; i++) x += (*p1++)*(*p2++); 
  return x;
}


void Vector ::print(){
  int i;
  
  for (i = 0; i< SIZEVECT; i++)
    printf("%f ",data[i]);

}

void Vector ::load(double x){
  int i;
  
  for (i = 0; i< SIZEVECT; i++)
    data[i] = x;

}

void Vector::setValue(int i, double x)
{
  data[i] = x;
}
double Vector::value(int i)
{
  return data[i];
}

Vector Vector::operator +(const Vector &rhs){
  Vector T;
  int i;
  T.size = this->size;
  // printf("(+)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = this->data[i] + rhs.data[i];
  return T;
}

Vector Vector::operator -(const Vector &rhs){
  Vector T;
  int i;
  T.size = this->size;
  // printf("(-)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = this->data[i] - rhs.data[i];
  return T;
}

Vector Vector::operator -(){
  Vector T;
  int i;
  T.size = this->size;
  // printf("(-u)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = -this->data[i];
  return T;
}

Vector Vector::operator /(const Vector &rhs){
  Vector T;
  int i;
  T.size = this->size;
  // printf("(/)%d\n", size);
  for(i = 0; i < size; i++)
    if(rhs.data[i] != 0.0)
       T.data[i] = this->data[i]/rhs.data[i];
     else printf("divide error %d\n");
  return T;
}


Vector Vector::operator *(const Vector &rhs){
  Vector T;
  int i;
  T.size = this->size;
  // printf("(*)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = this->data[i] * rhs.data[i];
  return T;
}

Vector &Vector::operator =(const Vector &rhs){
	int i;
        // printf("(=)%d\n", size);
	for(i = 0; i < rhs.size; i++)
		this->data[i] = rhs.data[i];
	return *this;
}

Vector &Vector::operator +=(const Vector &rhs)
{
  int i;
  // printf("(+=)%d\n", size);
  for(i = 0; i < rhs.size; i++)
    {
      this->data[i] += rhs.data[i];
    }
  return *this;
}

