#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h> /*BM*/

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
#ifndef INADDR_NONE
#  define INADDR_NONE ((u_long)0xffffffff)    /* -1 return */
#endif

#include "brkSockets.h"

#define DEBUG 0

int brk_server_connection(int serv_tcp_port) {
  int sockfd,newsockfd,clilen,childpid, tmp, waited;
  struct sockaddr_in cli_addr,serv_addr;
  /* IPPROTO_TCP  */
  if ((sockfd=socket(AF_INET,SOCK_STREAM,0))<0) {
    return brk_err_dump("server:can't open stream socket");
  }

  memset((char *) &serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
  serv_addr.sin_port = htons(serv_tcp_port);

  tmp = SOCKET_BUF_SIZE;
  
  /* no buffers; */
  if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, &tmp, sizeof (int)) == -1) {
    printf("setsockopt sndbuf serverconnection failed.\n");
  }
  if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &tmp, sizeof (int)) == -1) {
    printf("setsockopt rcvbuf serverconnection failed.\n");
  }

#if DEBUG
  printf("about to try to bind to socket %d\n", serv_tcp_port);
#endif

#define WAIT_TIME 5
  waited = 0;
  while(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0) {
    /* waiting for binding to come active; */
    fprintf(stderr, "server:can't bind local address, \n");
    fprintf(stderr, "\ttrying again in %d seconds.\n", WAIT_TIME);  
    sleep(WAIT_TIME);
    waited = 1;
  }

#if DEBUG
  printf("bound to socket %d\n", serv_tcp_port);
#endif

  if (waited) {
    printf("\n\tSocket bound.\n");
    fflush(stdout);
  } else {
#if DEBUG
    /* always print this message when DEBUG is true; */
    printf("\n\tSocket bound.\n");
    fflush(stdout);
#endif
  }
  
  listen(sockfd,5);


#if DEBUG
  printf("\n\tlistened at Socket.\n");
  fflush(stdout);
#endif

  clilen=sizeof(cli_addr);

  newsockfd=accept(sockfd,(struct sockaddr *) &cli_addr,&clilen);

#if DEBUG
  printf("\n\tConnection established.\n");
  fflush(stdout);
#endif

  if(newsockfd<0)
    return brk_err_dump("server:accept error");

  return newsockfd;
}

int brk_client_connection(char *serv_host_addr, int serv_tcp_port) {

  struct hostent *host; /*BM*/
  int sockfd, tmp, waited;
  struct sockaddr_in serv_addr;

  memset((char *) &serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_port=htons(serv_tcp_port);

  /*BM*/
  /* -- interpret serv_host_addr as ip address, if it fails -- */
  /* -- try interpreting as hostname                        -- */
  if ( (serv_addr.sin_addr.s_addr=inet_addr(serv_host_addr)) == INADDR_NONE ) {
    if ((host = gethostbyname(serv_host_addr)) == NULL) {
      fprintf(stderr, "client:can't get ip address from name %s\n",
              serv_host_addr);
      return -1;
    }
    memcpy((char*)&serv_addr.sin_addr, host->h_addr, host->h_length);

#if DEBUG
    fprintf (stderr, "%s translated to %s\n", serv_host_addr,
                                              inet_ntoa(serv_addr.sin_addr));
#endif
  }
  
  do {

    if((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
      return brk_err_sys("client:can't open stream socket");
    
    /* no buffers; */
    tmp = SOCKET_BUF_SIZE;
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, &tmp, sizeof (int)) == -1) {
      printf("setsockopt sndbuf client connection failed.\n");
    }
    
    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &tmp, sizeof (int)) == -1) {
      printf("setsockopt rcvbuf  client connection failed.\n");
    }
    
    if (waited == 1) {
      /* waiting for binding to come active; */
      fprintf(stderr, "client:can't connect to server, \n");
      fprintf(stderr, "\ttrying again in %d seconds.\n", WAIT_TIME);  
      sleep(WAIT_TIME);
      waited = 1;
    }
  } while (connect(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr))<0 
	   && (waited = 1));
    


  if (waited) {
    printf("\n\tSocket bound.\n");
    fflush(stdout);
  }

  return sockfd;
}

int brk_err_sys(char *str)
{
  fprintf(stderr,"%s: ",str);
  return -1;
}

int brk_err_dump(char *str)
{
  fprintf(stderr,"%s: ",str);
  fflush(stdout);
  fflush(stderr);
  return -1;

}

int
brk_sockWriten(register int fd, register char *ptr, register int nbytes)
{
  int sockII;
  int nleft,nwritten;
  nleft=nbytes;
  while (nleft >0)
    {
      nwritten=write(fd,ptr,nleft);
      if (nwritten<=0)
	return (nwritten);
      nleft-=nwritten;
      ptr += nwritten;
    }
  sockII = dup(fd);
  close(sockII);
  return (nbytes-nleft);
}

int
brk_sockWritestr(register int fd, register char *ptr)
{
  int nleft,nwritten, nbytes;

#if DEBUG
  printf("in sockWritestr writing string: %s,", ptr);
#endif

  nbytes = nleft = strlen(ptr)+1;
  while (nleft > 0) {
    nwritten=write(fd,ptr,nleft);
    if (nwritten<=0)
      return (nwritten);
    nleft-=nwritten;
    ptr += nwritten;
  }
#if DEBUG
  printf(" %d bytes\n", nbytes - nleft);
  fflush(stdout);
#endif
  return (nbytes - nleft);
}

int
brk_sockReadline(register int fd, register char *ptr, register int maxlen)
{
  int n,rc;
  char c;
  for (n=1;n<maxlen;n++)
    {
      if ((rc=read(fd,&c,1))==1)
	{
	  *ptr++ =c;
	  if (c=='\n')
	    break;
	}
      else if (rc==0)
	{
	  if (n==1)
	    return (0);
	  else
	    break;
	}
      else
	return (-1);
    }
  *ptr=0;
  return(n);
}


int
brk_sockReadlineFAST(register int fd, register char *ptr, register int maxlen)
{
  int n,rc;
  char c, ch[11];
  for (n=1;n<maxlen;n++)
    {
      if ((rc=read(fd,ch,10))>1)
	{
	  strncpy(ptr, ch, rc);
	  ptr += rc;
	  if (*ptr=='\n')
	    break;
	}
      else if (rc==0)
	{
	  if (n==1)
	    return (0);
	  else
	    break;
	}
      else
	return (-1);
    }
  *ptr=0;
  return(n);
}


int
brk_sockReadStrFast(register int fd, register char *ptr, register int maxlen) {
  int n,rc;
  char c, ch[11];
  for (n=1;n<maxlen;n++) {
    /* read in one char into 'c' */
    if ((rc=read(fd,ch,10))==1) {
      /* store read char 'c' into ptr string; */
      strncpy(ptr, ch, rc);
#if DEBUG
      printf("read in %s\n (%d)\n", ch, rc);
#endif
      ptr += rc;
      if (*ptr=='\n')
	break;
    } else 
    if (rc==0) {
      /* if nothing was read in; */
      if (n==1) {
	/*  if it was the first loop..return NULL; */
	return (0);
      }	else {
	/* it wasnt the first loop, return the number of bytes read; */
	break;
      }
    } else  {
      /* there was an error somewhere; */
      return (-1);
    }
  }
  /* add a null char onto end of ptr if not already there.; */
  if ((ptr-1) != NULL) {
    *ptr=0;
  }
  return(n);
}



int
brk_sockReadStr(register int fd, register char *ptr, register int maxlen) {
  int n,rc;
  char c;
  for (n=1;n<maxlen;n++) {
    /* read in one char into 'c'; */
    if ((rc=read(fd,&c,1))==1) {
      /* store read char 'c' into ptr string; */
      *ptr = c;
      /* and incr ptr; */
      ptr++;
      /* if c is null string, end this loop; */
      if (c=='\0')
	break;
    } else 
    if (rc==0) {
      /* if nothing was read in; */
      if (n==1) {
	/* if it was the first loop..return NULL; */
	return (0);
      }	else {
	/* it wasnt the first loop, return the number of bytes read; */
	   
	break;
      }
    } else  {
      /* there was an error somewhere; */
      return (-1);
    }
  }
  /* add a null char onto end of ptr if not already there.; */
  if ((ptr-1) != NULL) {
    *ptr=0;
  }
  return(n);
}
 
