# -----------------------------------------
#
#  Darryl Brown
#  April 1994
#
#  Socket communication for breakpoint executive;
#  
#
# -----------------------------------------


# -----------------------------------------
# Set up globals
set SocketNum 6871;
set ControlSocketNum 6872;
set SocketID 0;
set ControlSocketID 0;

set HostAddr 128.223.8.35

# Set up menu requests that can be made from 
# breakpoint server.
set serverMenu(list_all) 1
set serverMenu(get_types) 2
set serverMenu(set_continue) 3
set serverMenu(exit) 4


# -----------------------------------------
# initialize connection, assign socket 
# number.

set ConnectRole "server";

proc setUpConnection {} {
  global SocketID SocketNum HostAddr ConnectRole env;
  
  if {[info exist env(pcxx_BRKHOST)]} {
    set HostAddr $enf(pcxx_BRKHOST);
  } 

  if {[info exist env(pcxx_BRKCTRLSOCKET)]} {
    set ControlSocketNum $env(pcxx_BRKCTRLSOCKET);
  } 

  if {[info exist env(pcxx_BRKSOCKET)]} {
    set SocketNum $env(pcxx_BRKSOCKET);
  } 

  if {[string compare $ConnectRole "client"] == 0} {
    puts "setting client up connection.. $SocketNum $HostAddr";
    flush stdout;
    set SocketID [clientConn $SocketNum $HostAddr];
    puts "connection set.";
    flush stdout;
  } else {
    #must be server
    puts "setting server up connection.. $SocketNum $HostAddr";
    flush stdout;
    set SocketID [serverConn $SocketNum];
    puts "connection set.";
    flush stdout;
  }
}


# -----------------------------------------
proc sendMessage {mess} {
  global SocketID;
  
  if {$SocketID} {
    return [writeSocket $SocketID $mess];
  } else {
    return "ERROR no socket open";
  }
}  


# -----------------------------------------
proc menuChoice {choice {otherargs ""}} {
  global serverMenu;
  
  set ret [sendMessage $serverMenu($choice)];
  if {[string length $otherargs]} {
    return [sendMessage $otherargs];
  } else {
    return $ret;
  }
}


# -----------------------------------------
proc continueExec {} {
  sendMessage "";
}


# -----------------------------------------
proc getAnswer {} {
  global SocketID;
 
  set retstr "";
  set line [readAllSocket $SocketID];
  if {$SocketID} {
    set endFormat [format "\n!!\n"];
    while {[string compare $line $endFormat]} {
      # puts "retrieved :$line:";
      append retstr $line;
      set line [readAllSocket $SocketID];
    }
    # puts "final return string  :$retstr:";
    return $retstr;
  } else {
    return "ERROR no socket open";
  }
}

# -----------------------------------------
proc requestData {reqStr} {
  sendMessage $reqStr;
  return [getAnswer];
}

# -----------------------------------------
proc vizData {vizChannel collNum elemNumi elemNumj fname extract} {
  global SocketID;
  
  if {$extract==1} {
    set addendum " 1 ";
  } else {
    set addendum " ";
  }

  #puts stdout \
  # "collNum $collNum elemNumi $elemNumi elemNumj $elemNumj field $fname";

  if {[llength $elemNumj]} {
    set vcom [concat vizDataFromSocket $SocketID $vizChannel \
	    "($collNum,$elemNumi,$elemNumj)->$fname" $addendum];
    return [eval $vcom];
  } else {
    set vcom [concat vizDataFromSocket $SocketID $vizChannel \
	    "($collNum,$elemNumi)->$fname"  $addendum ];

    return [eval $vcom];
  }
}


# -----------------------------------------
proc getTypes {} {
  menuChoice get_types;
  return [getAnswer];
}


# -----------------------------------------
proc quitExecution {} {
  menuChoice exit;
  return;
}


# -----------------------------------------
proc getCollElem {coll elemi elemj} {

  if {[llength $elemj]} {
    # puts stdout "getting collelem ($coll,$elemi,$elemj)";
    return [requestData "($coll,$elemi,$elemj)"];
  } else {
    # puts stdout "getting collelem ($coll,$elemi)";
    return [requestData "($coll,$elemi)"];
  }
}
  


# -----------------------------------------
proc getCollElemField {coll elemi elemj field} {
  if {[llength $elemj]} {
    # puts stdout "getting collelem ($coll,$elemi,$elemj)->$field";
    return [requestData "($coll,$elemi,$elemj)->$field"];
  } else {
    # puts stdout "getting collelem ($coll,$elemi)->$field";
    return [requestData "($coll,$elemi)->$field"];
  }
}
  

# -----------------------------------------
proc skipBreakpoints {skipNum} {
  append rexp {[^} "0-9" {]};
  # if I cant find any non-digits, continue
  if {![regexp $rexp $skipNum]} {
    menuChoice set_continue $skipNum;
    return $skipNum;
  } else {
    return -1;
  }
}


# -----------------------------------------
proc contBreakpoints {contNum} {
  puts "in contBreakpoints";
  flush stdout;
  append rexp {[^} "0-9" {]};
  # if I cant find any non-digits, continue
  if {![regexp $rexp $contNum]} {
    for {set i 0} {$i < $contNum} {incr i} {
      continueExec;
      puts "getting instances";
      flush stdout;
      getInstances;
      puts "updating idletasks";
      flush stdout;
      update;
      update idletasks;
      puts "done updating idletasks";
      flush stdout;
    }
    puts "leaving contBreakpoints";
    flush stdout;
    return $contNum;
  } else {
    puts "leaving contBreakpoints";
    flush stdout;
    return -1;
  }
}

# -----------------------------------------
proc getInstances {} {
  global InstanceList;
  
  set InstanceList "";
  menuChoice list_all;

  puts "getting instances";
  flush stdout;

  set strPointer [getAnswer];

  if [regexp "^id=(.*)#fn=(.*)#ln=(.*)#(.*)$" $strPointer dummy \
      funcId fileName lineNumber strPointer] {
    if { $funcId != -1 } {
      #puts "stopped in $funcId called from $fileName:$lineNumber";
      #flush stdout;
      global_selectLine $lineNumber $fileName $funcId
    }
  } else {
    puts "communication problem: no funcId";
  }
  set rexp [format "^\[^\n\]*\n"];
  
  puts "parsing instances";
  flush stdout;

  while {[regexp -indices $rexp $strPointer ind]} {
    set tline [string range $strPointer [lindex $ind 0] [lindex $ind 1]];
    set strPointer [string range $strPointer [expr [lindex $ind 1]+2] end] ;
    if {[scan $tline { %s < %s > %s %s} collName elemName numElems varName] \
	  == 4} {
      set line "$collName < $elemName > $numElems $varName";
      lappend InstanceList [list $line];
    } else {
      puts "communication problem .$tline.";
    }
  }
  
  puts "autoupdating";
  flush stdout;

  autoUpdate;

  puts "done autoupdating";
  flush stdout;

  return $InstanceList;
}


# -----------------------------------------
set autoUpdateList "";

proc addAutoUpdate {tw collType elemRange field} {
  global autoUpdateList ;
  
  set found 0;

  foreach elem $autoUpdateList {
    if {[string compare [lindex $elem 0] $tw] == 0} {
      set found 1;
    }
  }
  if {!$found} {
    lappend autoUpdateList [list $tw $collType $elemRange $field];
  }
}

proc delAutoUpdate {tw} {
  global autoUpdateList ;

  set newlist "";

  foreach elem $autoUpdateList {
    if {[string compare [lindex $elem 0] $tw] == 0} {
      # dont append;
    } else {
      lappend newlist $elem;
    }
  }
  set autoUpdateList $newlist;
}


proc delAutoUpdateViz {vizwin} {
  delAutoUpdate $vizwin;
}

proc addAutoUpdateViz {vizType collType elemRange field vizwin} {
  global autoUpdateList ;
  
  set found 0;

  foreach elem $autoUpdateList {
    if {[string compare [lindex $elem 0] $vizwin] == 0} {
      set found 1;
    }
  }
  if {!$found} {
    lappend autoUpdateList [list $vizwin VIZ $collType $elemRange \
			    $field $vizType];
  }
}

proc autoUpdate {} {
  global autoUpdateList Clist;

  set delList {};


  foreach elem $autoUpdateList {
    #check to see if this is a vizualization;
    # puts stdout "updateing using .$elem. element fo update list";

    if {![strsame [lindex $elem 1] VIZ]} {
      # NOT a visualization.
      set tw [lindex $elem 0];
      if {![winfo exist $tw]} {
	lappend delList $tw;
      } else {
	set collType [lindex $elem 1];
	set numDims [getNumDims $collType];
	set coords [lindex $elem 2];
	set field [lindex $elem 3];
	

	# puts "\"$elem\", ct $collType nd $numDims c $coords";

	if {([llength $coords] >1) && ([llength [lindex $coords 1]] == 0)} {
	  # puts "setting coords from $coords to [lindex $coords 0]";
	  set coords [lindex $coords 0];
	}

	if {($numDims < 2) || (($numDims < 2) && ([llength $field] == 0))} {
	  set ux [lindex $coords 0];
	  set uy [lindex $coords 1];
	} else {
	  set ulcoords [lindex $coords 0];
	  set lrcoords [lindex $coords 1];
	  set ux [lindex $ulcoords 0];
	  set uy [lindex $ulcoords 1];
	  set lx [lindex $lrcoords 0];
	  set ly [lindex $lrcoords 1];
	}
	
	
	for {set collNum 0}  {![strsame $collType [$Clist get $collNum]] && \
				$collNum < [$Clist size]} \
				  {incr collNum} {
				  };
	if {$collNum >= [$Clist size]}  {
	  puts stderr "No insntance of type $collType";
	  return ;
	}
	
	$tw delete 1.0 end;

	if {[string length $field]} {
          #must be something in the field
	  set retrievComm [concat "showFieldWin " \
			   {[getData $collNum $i $j $field $collType]} \
			   $field [list $collType] $tw;];

	  if {($numDims > 1)} {
	    for {set i $ux} {$i <= $lx} {incr i} {
	      for {set j $uy} {$j <= $ly} {incr j} {
		$tw insert end [eval $retrievComm];
	      }
	    }
	  } else {
	    set j "";
	    for {set i $ux} {$i <= $uy} {incr i} {
	      $tw insert end [eval $retrievComm];
	    }
	  }
	} else {
          #must NOT be something in the field;
	  set retrievComm [concat "displayDataWin" \
			   {[getData $collNum $ux $uy "" $collType]} \
			   [list $collType] $tw ;];
	  $tw insert end [eval $retrievComm];
	}
	
	
      }
    } else {
      # IT IS a vizualization;
      set vizwin [lindex $elem 0];

      set collType [lindex $elem 2];
      set numDims [getNumDims $collType];
      set coords [lindex $elem 3];
      # puts stdout "elem <$elem>";

      if {$numDims < 2} {
	set ux [lindex $coords 0];
	set uy [lindex $coords 1];
	# puts stdout "ux uy $ux $uy";
      } else {
	set ulcoords [lindex $coords 0];
	set lrcoords [lindex $coords 1];
	set ux [lindex $ulcoords 0];
	set uy [lindex $ulcoords 1];
	set lx [lindex $lrcoords 0];
	set ly [lindex $lrcoords 1];
	# puts stdout "ul lr $ulcoords $lrcoords ux uy $ux $uy lx ly $lx $ly coords $coords";
      }
      
      
      set field [lindex $elem 4];
      set dtype [findFType [extractElem $collType] $field];

      if {![isArray $dtype]} {
	set addendum 1;
      } else  {
	set addendum 0;
      }

      # puts stdout "addendum $addendum dtype $dtype collType $collType";
      set vizType [lindex $elem 5];
      for {set collNum 0}  {![strsame $collType [$Clist get $collNum]] && \
			      $collNum < [$Clist size]} \
				{incr collNum} {
				};
      if {$collNum >= [$Clist size]}  {
	puts stderr "No insntance of type $collType";
	return ;
      }
      if {![llength [set viz [alreadyActive $vizwin]]]} {
	puts stderr "No viz for $collType : $vizType -> $field";
	return;
      } else {
	set vizChannel [lindex $viz 1];
      }
      
      puts "calling vizdata command...";
      flush stdout;

      if {$numDims > 1} {
	set vCommand [concat vizData $vizChannel $collNum \
		      {$i $j} $field $addendum];
	for {set i $ux} {$i <= $lx} {incr i} {
	  for {set j $uy} {$j <= $ly} {incr j} {
	    eval $vCommand;
	  }
	}
      } else {
	set vCommand [concat vizData $vizChannel $collNum \
		      {$i} {""} $field $addendum];
	#puts stdout "vcommand $vCommand"
	for {set i $ux} {$i <= $uy} {incr i} {
	  eval $vCommand;
	}
      }

      puts "flushing viz channel...";
      flush stdout;

      flush $vizChannel;
    }
  }

  puts "almost done autoupdating...";
  flush stdout;

  foreach tw $delList {
    delAutoUpdate $tw;
  }

  puts "finally  done autoupdating..."
    flush stdout;
}

  
