// This may look like C code, but it is really -*- C++ -*-

#ifndef ASGFUNCS3_H
#define ASGFUNCS3_H 1

#ifdef IN_ASGFUNCS3_C
#define EXTERN
#else
#define EXTERN extern
#endif



/*
  isConstantExpr:
  
  I want to know that (10+5)*3 is a constant expression. This function
  will tell me that.

  The function `isConstantExpr' checks the expression tree it gets for
  values, casts, arithmetic ops, EXPR_LIST nodes, and PAREN_OP nodes.
  If it finds something that is not one of those, it believes that the
  expression is non-constant, an returns false. If the whole tree is
  this type of node, it returns true.

  WARNING: This function considers any cast of a non-floating type
  to a floating type to be CONSTANT! This is because such an object
  will have no gradient.

  To be 100% safe using this function, interpret "constant" to me
  "is known to have no gradient".

  */

EXTERN int 
isConstantExpr (SgExpression *the_expr);

/*
  `isSafeExpr' checks an expression to see if it is "safe"
  to evaluate more than once. One very nice thing this will do is
  let us just repeat lvalues that appear, rather than worrying that
  they might be the result of function calls.
  
  So far, the following things are safe:
       values, casts, sizeof ops, arithmetic operations, EXPR_LIST nodes,
       PAREN_OP nodes, variable references, array references,
       POINTST_OP (->), RECORD_REF (.), logical operations,
       bit shift operations, address-of operations, mod ops,
       and dereferences(!)

  We begin with the presumption that the expression is safe, and
  flag it as unsafe if we find anything bad.

 */
EXTERN int
isSafeExpr (SgExpression *the_expr);


/*
  classifyCast:

  Determine whether a cast is:
  () Cast of anything to a non-floating type.
  () Cast of a non-floating type to a floating type.
  () Cast of floating type to the same floating type. (No effect cast.)
  () Cast of one floating type to another floating type.
  [This includes cast of float to unsigned float and the like.]

  or:
  () Not a cast.
  () Cast with no children. [Appears as operand of `sizeof'.]

  */

typedef enum Cast_Classification {
     NOT_A_CAST = 0,
     SOMETHING_TO_NON_FLOATING,
     NON_FLOATING_TO_FLOATING,
     FLOATING_TO_SAME_FLOATING,
     FLOATING_TO_OTHER_FLOATING,
     NO_CHILDREN
} Cast_Classification;

EXTERN Cast_Classification
classifyCast (SgExpression *the_expr);



#undef EXTERN
#endif /* ndef ASGFUNCS3_H */
