// This may look like C code, but it is really -*- C++ -*-

#include "sage++user.h"
#include "iterator-expr.h"
#include <stdlib.h>
#include "error.h"

void
Iterator_Expr::begin (void)
{
     if ( our_stmt != 0 )
     {
	  /* Process all expressions in a statement */

	  for (int i = 0; i < 3 ; i++ )
	  {
	       SgExpression *replacement;
     
	       SgExpression *current_expr = our_stmt->expr(i);
	       this->current_depth = 0;

	       if ( current_expr != 0 )
	       {
		    replacement = iterate (current_expr);
		    our_stmt->setExpression(i,replacement);
	       }
	  }
     }

     if ( our_expr != 0 )
     {
	  this->current_depth = 0;
	  iterate (our_expr);
     }

}


/*
  The CURRENT_DEPTH points at the one past the last index assigned to
  in the `depth' array (stack). The valid range of the depth stack is
  depth[0] ... depth[depth_allocated-1].

  The item returned by `iterate' is supposed to replace the
  CURRENT_EXPR it got.
  
 */

SgExpression *
Iterator_Expr::iterate (SgExpression *current_expr)
{
     SgExpression *current_replacement;

     if ( current_depth >= depth_allocated )
     {
	  depth_allocated += ITERATOR_EXPR_DEPTH_INCREMENT;
	  depth = (SgExpression **) realloc ( depth,
					      depth_allocated
					      * sizeof (SgExpression *));
	  if ( depth == 0 )
	       fatal_error ("Out of memory in Iterator_Expr reallocation.");
     }

     depth[current_depth] = current_expr;
     current_depth++;

     /* The pointer returned by `process' should replace the CURRENT_EXPR
	as the item pointed to by the parent, unless they are the same
	or the returned item is zero (in which case the descent
	should stop). */
     
     current_replacement = process (current_expr);

     if ( current_replacement != 0 )
     {
	  SgExpression *lower_replacement;

	  if (current_expr->lhs() != 0)
	  {
	       lower_replacement = iterate(current_expr->lhs());
	       current_expr->setLhs(lower_replacement);
	  }
	  
	  if (current_expr->rhs() != 0)
	  {
	       lower_replacement = iterate(current_expr->rhs());
	       current_expr->setRhs(lower_replacement);
	  }
     }

     current_depth--;

     if ( current_replacement != 0 )
	  return current_replacement;
     else
	  return current_expr;
}


void
Iterator_Expr::zero_current (void)
{
     SgExpression *current = this->parent(0);
     SgExpression *the_parent = this->parent(1);

     SgExpression *null_expr = 0;
     
     if ( the_parent->lhs() == current )
	  the_parent->setLhs(null_expr);

     else if ( the_parent->rhs() == current )
	  the_parent->setRhs(null_expr);

     else
	  fatal_error ("Unable to find child in `zero_current'\n");
     
}
