// This may look like C code, but it is really -*- C++ -*-

#ifndef ITERATOR_EXPR_H
#define ITERATOR_EXPR_H 1

#include "sage++user.h"

/* This is the deepest the recursive traversal can go. */

#define ITERATOR_EXPR_DEPTH_INCREMENT 10

class Iterator_Expr {
  public:
     Iterator_Expr (SgStatement *the_stmt);
     Iterator_Expr (SgExpression *the_expr);     

     virtual ~Iterator_Expr();

     /* This begins the iteration */
     void begin (void);

     /* The parent function returns the i-th parent of the current 
        expression. The immediate parent of the current expression
	may be retrieved as parent(1) or parent(). */
     SgExpression *parent (int i = 1);

     /* The stmt function returns the associated statement, if any. */
     SgStatement *stmt(void);

     /*
        `process':

        This member function is called for every expression in the
        scope of the iteration. Subclasses must override this! 

	The current expression should be returned if no special behaviour
	is desired.
            o If a different expression is returned, the current
	      expression will be replaced by that expression.
	    o If a 0 expression	pointer is returned, the iteration
	      will not descend any children of that expression.

	In order to zero out the parent's pointer to the current expression
	the user should call `zero_current' in the user's process code.
	*/

     virtual SgExpression *process (SgExpression *current_expr) = 0;

     /*
       `zero_current':
       
       This sets the parent's pointer to the current child to 0.
       It requires that expression pointers be unique. [Should be true
       unless you do dangerous things in your code.]
       */
     
     void zero_current (void);
     
     int current_depth;
     int depth_allocated;
     SgExpression **depth;

  private:
     SgExpression *iterate (SgExpression *current_expr);

     SgStatement *our_stmt;
     SgExpression *our_expr;

};

#include "iterator-expr.icc"

#endif /* ndef ITERATOR_EXPR_H */
