// This may look like C code, but it is really -*- C++ -*-

#ifndef ITERATOR_H
#define ITERATOR_H 1

#include "sage++user.h"
#include "Symbol_Maintainer.h"

class Iterator {
  public:
     Iterator (SgFile *file);
     virtual ~Iterator();

     /* This begins the iteration */
     void begin (void);

     /* This member function is called for every statement in the
	scope of the iteration. Subclasses must override this! */
     virtual void process (SgStatement *current_stmt) = 0;

     /* Variables that indicate where we are in the traversal */
     SgStatement *current;
     SgStatement *next;

     /************ Symbol Maintenance *************/
       
     /* These are used by the iterator, and may also be used
	by your "process" routine, to manage symbol lookup. */

     /* Note: The SOFT_AND_HARD member of the Scope_Movement enum must
	be the bitwise OR of HARD and SOFT. Cheesy programming. */
     
     enum Scope_Movement { HARD = 1, SOFT = 2,
			   HARD_AND_SOFT = 3, SOFT_AND_HARD = 3 };

     virtual void down_namescope (enum Scope_Movement which = SOFT_AND_HARD);
     virtual void up_namescope (enum Scope_Movement which = SOFT_AND_HARD);

     Symbol_Maintainer *namescope;
     Symbol_Maintainer *hard_namescope;

  protected:
     /* I can't think of why you would want to override this,
	but just in case, I'll keep it merely `protected'. */
     
     void iterate (SgStatement *current_stmt);

  private:
     SgFile *the_file;
};

#include "iterator.icc"

#endif /* ndef ITERATOR_H */
