// This may look like C code, but it is really -*- C++ -*-

#include <stdio.h>
#include "sage++user.h"
#include "iterator.h"

void
Iterator::begin (void)
{
     SgStatement *first_stmt = the_file->firstStatement();

     /* Init vars. */
     prev = 0;
     next = 0;
     current = first_stmt;
     
     before_descent (first_stmt);

     /* Do stuff */
     iterate (first_stmt);

     after_ascent (first_stmt);
}



/*
  Iterator::iterate

  One stupid thing that I want to mention: We keep track of the current
  and next statement in local variables so we can call ourselves recursively.
  However, other routines may want this information, so at various points
  I copy the local information back to the class-"global"
  vars: "prev," "current," and "next". 

  If this function returns -1, this is an indication that the parent
  should reprocess the current child list because a "back up" request
  was received from the user's `process' function.
  
 */

#define UPDATE_CURRENT() \
     prev = prev_stmt; current = current_stmt; next = next_stmt;

int
Iterator::iterate (SgStatement *current_stmt)
{
     int user_action;
     int back_up;

     SgStatement *next_stmt = 0;
     SgStatement *prev_stmt = 0;
     
     while ( current_stmt != NULL )
     {
          /* We have to find the next statement first thing, because
             one of the actions we may want to perform is deleting 
             the current statment! Needless to say, this screws 
             up our ability to find the next statement later. */

          next_stmt=current_stmt->nextInChildList();

          /* Update the current and next members */
          UPDATE_CURRENT();

          user_action = process (current_stmt);

	  if ( user_action == 0 )
	  {
	       /* Do not descend children */
	  }
          else if ( user_action == 1 )
          {
               /* Descend the children of the current statement, if it
                  has any. */

	       int sub_user_action;
          
               /* Child List 1 */

               if (current_stmt -> numberOfChildrenList1() > 0)
               {
                    before_descent(current_stmt);

		    do { 
			 sub_user_action =
			      iterate (current_stmt->childList1(0));
		    } while (sub_user_action == -1);

                    /* Restore the current and next members */
                    UPDATE_CURRENT();
                    after_ascent(current_stmt);
               }

               /* Child List 2 */
               
               if (current_stmt -> numberOfChildrenList2() > 0)
               {
                    before_descent (current_stmt);

		    do { 
			 sub_user_action =
			      iterate (current_stmt->childList2(0));
		    } while (sub_user_action == -1);
                    
                    /* Restore the current and next members */
                    UPDATE_CURRENT();
                    after_ascent (current_stmt);
               }
          }
          else if ( user_action == -1 )
	  {
	       /* If we need to go back and repeat something, do it now. */
	       if ( prev_stmt == 0 )
	       {
		    /* We are the first in our child list; we need
		       to go back up to the parent to repeat */
		    return -1;
	       }
	       else
	       {
		    /* No need to go back up to the parent to repeat.
		       Adjust stuff here so that at the end, when we move
		       along the list, we rescan what we just did.

		       Here we set current = previous, and then
		       recomput the next ("the stmt after the one before
		       the statement we just processed"). */

		    current_stmt = prev_stmt;
		    next_stmt = prev_stmt->nextInChildList();
	       }
	  }
	  else
	  {
	       /* Some kind of check seems like a good idea since I
		  changed the return type. */
	       fprintf(stderr,"Bad programmer usage of Iterator class\n");
	       abort();
	  }

          /* Then continue along the current childlist */
          prev_stmt = current_stmt;
          current_stmt = next_stmt;

     } /* end while (current_stmt != NULL) */
}

