#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#

#
# initDep: initialize / reset global variables
#

proc initDep {} {
  global depfile

  set depfile(host) "localhost"; # hostname of machine where depfile resides
  set depfile(arch) "???";   # architecture of machine where depfile resides
  set depfile(hostarch) "";  # hostname + architecture
  set depfile(root) "???";   # SAGEROOT of machine where depfile resides
  set depfile(path) "";      # pathname of depfile
  set depfile(dir)  "";      # directory part of pathname or '.'
  set depfile(file) "";      # filename part of pathname
  set depfile(maintag) "";   # Sage++ id for pC++ main function
  set depfile(allftags) "";  # the ids of all functions
  set depfile(allctags) "";  # the ids of all classes
  set depfile(numfunc) 0;    # number of functions defined in depfile
  set depfile(numused) 0;    # number of functions used (called) in program
  set depfile(numprof) "";   # number of functions profiled
  set depfile(ino) -1;       # inode of depfile
}

#
# readDep: read depfile by using "cgm"
#
#  option: option to use for cgm
#    path: UNIX pathname of depfile to read
#

proc readDep {option path} {
  # -- get access to global vars
  global depfile
  global funcs
  global classes members
  global BINDIR REMBINDIR REMSH

  # -- check file
  if { $path == [status RR] } {
    showError "Cannot reload file: `$path'." fatal
    return NOT_OK
  } elseif { $depfile(host) == "localhost" } {
    # -- depfile on local host
    if { ! [string match *.dep $path] } {
      showError "No dep file: `$path'."
      return NOT_OK
    } elseif { ! [file readable $path] } {
      showError "Cannot open file: `$path'."
      return NOT_OK
    }
    file stat $path statbuf
    set depfile(ino) $statbuf(ino)
  } else {
    # -- depfile on remote host
    set depfile(ino) -1
  }
  SetDepfileBasics $path

  if { $depfile(host) == "localhost" } {
    # -- change directory to the one where depfile is located
    # -- so that relative pathnames coming from Sage++ are
    # -- interpreted correctly
    cd $depfile(dir)

    # -- read directly from cgm
    set readcom "|$BINDIR/cgm $option $depfile(file)"
  } else {
    set REMBINDIR "$depfile(root)/bin/$depfile(arch)"

    # -- read through remote shell from cgm
    set readcom "|$REMSH $depfile(host) \
                -n \"cd $depfile(dir); $REMBINDIR/cgm $option $depfile(file)\""
  }

  # -- initialize global variables
  set numfunc 0

  # -- open access to cgm reader
  set in [open $readcom r]

  # -- output for functions have header containing number of functions
  if { $option != "-dumpch" } {
    # -- Read number of functions
    gets $in line
    scan $line "%d" depfile(numfunc)
    gets $in line
  }

  # -- Read data from cgm
  while {[gets $in line] >= 0} {
    scan $line "%s %s %s %s" com arg1 arg2 arg3
    switch $com {
      ftag:   { # -- tag: <tag> <name>
                set tag $arg1
                lappend depfile(allftags) $tag
                set funcs($tag,state) 0
                set funcs($tag,fold) 1
                incr numfunc

                set funcs($tag,name) $arg2
                if { $arg2 == "Processor_Main" } {set depfile(maintag) $tag}
              }

      fclass: { # -- class: <name|-> <type|-> <tag|-2>
                set funcs($tag,class) [list $arg1 $arg2 $arg3]
              }

      ffile:  { # -- file: <line|0> <file|-> @ <position>
                set funcs($tag,file) [list $arg1 $arg2]

                if { $arg1 != 0 } {
                  set x [string first "@" $line]
                  incr x 2
                  set funcs($tag,pos) [string range $line $x end]
                }
              }

      ftype:  { # -- type: <par|seq> <type> <used|not>
                set funcs($tag,type) [list $arg1 $arg2 $arg3]
                if { $arg3 == "used" } { incr depfile(numused) }
              }

      fcalls: { # -- calls: <num-called> - <list of calls>
                set funcs($tag,num) $arg1

                set x [string first "-" $line]
                incr x 2
                set funcs($tag,calls) [string range $line $x end]
              }

      fpos:   { # -- pos: <list of call-site-positions>
                set x [string first ":" $line]
                incr x 2
                set funcs($tag,childpos) [string range $line $x end]
              }

      fline:  { # -- line: <list of call-site-linenumbers>
                set x [string first ":" $line]
                incr x 2
                set funcs($tag,childline) [string range $line $x end]
              }

      ctag:   { # -- tag: <tag> <name> <COLL|->
                set tag $arg1
                if [ info exists classes($tag,name) ] {
                  set tag -2
                } else {
                  lappend depfile(allctags) $tag
                  set classes($tag,name) $arg2
                  set classes($tag,fold) 0
                  set classes($tag,coll) $arg3
                }
              }

      cfile:  { # -- file: <line|0> <file|-> @ <position>
                set classes($tag,file) [list $arg1 $arg2]

                if { $arg1 != 0 } {
                  set x [string first "@" $line]
                  incr x 2
                  set classes($tag,pos) [string range $line $x end]
                }
              }

      cpos:   { # -- pos: <list of type-positions>
                set x [string first ":" $line]
                incr x 2
                set classes($tag,typepos) [string range $line $x end]
              }

      csub:   { # -- sub: <list-of-subclasses>
                set x [string first ":" $line]
                incr x 2
                set classes($tag,subs) [string range $line $x end]
              }
      cmpos:  { # -- mpos: <list of member-func-decl-positions>
                set x [string first ":" $line]
                incr x 2
                set classes($tag,fmempos) [string range $line $x end]
              }

      cbase:  { # -- base: <list-of-baseclasses-and-positions>
                set x [string first ":" $line]
                incr x 2
                set classes($tag,basepos) [string range $line $x end]
              }

      cmem:   { # -- mem: <member> <tag> <type>
                set members($arg1) [lrange $line 1 3]
                lappend classes($tag,mem) $arg1
              }
    }
    set line ""
    set com ""
  }
  if [catch {close $in} errmsg] {
    showError "$readcom: `$errmsg'."
    return NOT_OK
  }
  return $numfunc
}

#
# findDep: try to determine depfile filename automatically
#          if only directory is given
#
#       f: input pathname
#

proc findDep {f} {
  global depfile
  global myname

  set pcxxsrc 0

  if { ! [regexp {(.*):(.*)} $f dummy depfile(host) path] } {
    # -- depfile on local host
    set depfile(host) "localhost"

    if { $f == [status RR] } {
      showError "Cannot reload file: `$f'." fatal
      return NOT_OK
    } elseif { ! [string match *.dep $f] && ! [ file isdirectory $f ] } {
      showError "No dep file or directory: `$f'."
      exit
    } elseif [ file isdirectory $f ] {
      # -- f is directory
      if [ file readable $f/SOURCE ] {
        # -- if there is a file named "SOURCE", take first name
        set in [open $f/SOURCE r]
        gets $in d
        close $in
        set dfile $f/$d.dep
        set pcxxsrc 1
      } else {
        # -- otherwise look if there is only one depfile
        set d [glob -nocomplain $f/*.dep]
        if { [ llength $d] == 1 } {
          set dfile $d
        } else {
          # -- no SOURCE file or more than one depfile
          showError "Cannot find depfile in `$f'.\n(no pC++ source directory)"
          exit
        }
      }
    } else {
      # -- no directory, so just take user-supplied filename
      set dfile $f
    }

    if { ! [file readable $dfile] } {
      if {!(($myname(name) == "cosy" || $myname(name) == "tau") && $pcxxsrc)} {
        showError "Cannot open file: `$dfile'."
      }
      return NOT_OK
    } else {
      return $dfile
    }
  } else {
    # -- remote host: if not ending in .dep, assume it is directory
    if [string match *.dep $path] {
      set dfile $path
    } else {
      set s [getOutput "cat $path/SOURCE"]
      if { $s == "NOT_OK" } {
        showError \
   "Cannot find depfile in `$depfile(host):$path'.\n(no pC++ source directory)"
        exit
      } else {
        set dfile $path/$s.dep
        set pcxxsrc 1
      }
    }

    if { [getOutput "ls $dfile"] == "NOT_OK" } {
      if {!(($myname(name) == "cosy" || $myname(name) == "tau") && $pcxxsrc)} {
        showError "Cannot open file: `$dfile'."
      }
      return NOT_OK
    } else {
      return $dfile
    }
  }
}

# -- initialize global variables
initDep
