#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#

source "$TAUDIR/inc/depfile.tcl"

set showFile(name) "";    # filename of currently selected function/class
set showFile(tag)  NONE;  # Sage++ id of currently selected function/class
set showFile(what) NONE;  # type of selected item

#
# selectLine: implementation of global feature "selectLine" for fancy
#             highlight specified line in file
#
#       line: line number to select
#       file: file name to select
#        tag: tag of call site to select or -1
#

proc selectLine {line file {tag -1}} {
  showFile $tag Line $file $line
}

#
# selectFuncTag: implementation of global feature "selectFunction" for fancy
#                display body of selected function in text window
#
#           tag: id of function to select
#

proc selectFuncTag {tag} {
  global funcs

  if [ info exists funcs($tag,name) ] {
    showFile $tag Func
    processTag $tag
  }
}

#
# showLine: highlight specified line in file
#
#      win: pathname of text window
#      tag: Sage++ id of call site
#     file: filename where source code of function can be found
#     line: linenumber where source code of function can be found
#

proc showLine {win tag file line} {
  global funcs

  $win tag add line $line.0 [expr $line+1].0
  $win tag configure line -background yellow -relief raised
  $win yview -pickplace $line.0

  if { $tag != -1 } {
    set name $funcs($tag,name)

    set text [$win get $line.0 [expr $line+1].0]
    if [regexp -indices $name $text range] {
      set cstart $line.[lindex $range 0]
      set cend   $line.[expr [lindex $range 1]+1]

      if [ lindex $funcs($tag,file) 0 ] {
        # -- child with body; add bindings to select child
        $win tag add child1 $cstart $cend
        $win tag configure child1 -background lightblue -relief raised
        $win tag bind child1 <Button-1> "global_selectFuncTag $tag"
        $win tag bind child1 <Button-2> "global_showFuncTag $tag"
        $win tag bind child1 <Button-3> "global_selectFuncTag $tag"
      } else {
        # -- library function (no body available); don't set up bindings
        $win tag add child1 $cstart $cend
        $win tag configure child1 -background orchid -relief raised
      }
    }
  }
}

#
# showFunc: highlight header, body, and children of selected function
#           in text window
#
#      win: pathname of text window
#      tag: Sage++ id of selected function
#     file: filename where source code of function can be found
#     line: linenumber where source code of function can be found
#

proc showFunc {win tag file line} {
  global funcs classes

  if { [scan $funcs($tag,pos) "%s %s %s %s" hstart hend bstart bend] == 4 } {
    # -- highlight header and body
    $win tag add func $hstart $hend
    $win tag configure func -background orange -relief raised
    $win tag add body $bstart $bend
    $win tag configure body -background yellow -relief raised
    $win yview -pickplace $line.0

    # -- if it is a member function, add bindings to
    # -- select its class
    set class [lindex $funcs($tag,class) 0]
    if { $class != "-" } {
      set ctag $classes($class,tag)
      $win tag bind func <Button-1> "selectClassTag $ctag"
      $win tag bind func <Button-3> "global_selectClassTag $ctag"
    }

    # -- highlight children (function calls within the body)
    set calls $funcs($tag,calls)
    set pos   $funcs($tag,childpos)
    set i 0
    foreach c $pos {
      scan $c "%s %s" cstart cend
      if { $cstart != "-" } {
        set ctag [lindex $calls $i]
        if [ lindex $funcs($ctag,file) 0 ] {
          # -- children with body; add bindings to select child
          $win tag add child$i $cstart $cend
          $win tag configure child$i -background lightblue -relief raised
          $win tag bind child$i <Button-1> "global_selectFuncTag $ctag"
          $win tag bind child$i <Button-2> "global_showFuncTag $ctag"
          $win tag bind child$i <Button-3> "global_selectFuncTag $ctag"
        } else {
          # -- library function (no body available); don't set up bindings
          $win tag add child$i $cstart $cend
          $win tag configure child$i -background orchid -relief raised
        }
      }
      incr i
    }
  }
}

#
# showClass: highlight header, body, and members of selected class
#            in text window
#
#      win: pathname of text window
#      tag: Sage++ id of selected class
#     file: filename where source code of class definition can be found
#     line: linenumber where source code of class definition can be found
#

proc showClass {win tag file line} {
  global classes funcs

  if { [scan $classes($tag,pos) "%s %s %s %s" hstart hend bstart bend] == 4 } {
    # -- highlight header and body
    $win tag add func $hstart $hend
    $win tag configure func -background orange -relief raised
    $win tag add body $bstart $bend
    $win tag configure body -background yellow -relief raised
    $win yview -pickplace $line.0

    # -- highlight attribute markers like public, private, ...
    set pos $classes($tag,typepos)
    foreach c $pos {
      scan $c "%s %s" cstart cend
      if { $cstart != "-" } {
        $win tag add ctype $cstart $cend
      }
    }
    $win tag configure ctype -background orange -relief raised

    # -- highlight function members
    set i 0
    set pos $classes($tag,fmempos)
    foreach c $pos {
      scan $c "%d %s %s" ctag cstart cend
      if { $cstart != "-" } {
        if [ lindex $funcs($ctag,file) 0 ] {
          $win tag add child$i $cstart $cend
          $win tag configure child$i -background lightblue -relief raised
          $win tag bind child$i <Button-1> "global_selectFuncTag $ctag"
          $win tag bind child$i <Button-2> "global_showFuncTag $ctag"
          $win tag bind child$i <Button-3> "global_selectFuncTag $ctag"
        } else {
          $win tag add child$i $cstart $cend
          $win tag configure child$i -background orchid -relief raised
        }
      }
      incr i
    }

    # -- add bindings to select base classes if necessary
    set i 0
    set pos $classes($tag,basepos)
    foreach c $pos {
      scan $c "%d %s %s" ctag cstart cend
      if { $cstart != "-" } {
        $win tag add base$i $cstart $cend
        $win tag bind base$i <Button-1> "selectClassTag $ctag"
        $win tag bind base$i <Button-3> "global_selectClassTag $ctag"
      }
      incr i
    }
  }
}

#
# markAllFuncAndClass: setup bindings for selecting all functions or classes
#                      in a file
#
#                 win: pathname of text window
#                file: filename to look for
#

proc markAllFuncAndClass {win file} {
  global funcs classes
  global depfile

  # -- look through all functions whether they are declared in the
  # -- specified file and setup bindings, if found one
  foreach t $depfile(allftags) {
    set f [lindex $funcs($t,file) 1]
    if { $f != "-" && [string compare $f $file] == 0 } {
      if { [scan $funcs($t,pos) "%s %s %s %s" hstart hend bstart bend] == 4 } {
        $win tag add func$t $hstart $bend
        $win tag configure func$t -background Gray95 -relief raised
        $win tag bind func$t <Button-1> "global_selectFuncTag $t"
        $win tag bind func$t <Button-2> "global_showFuncTag $t"
        $win tag bind func$t <Button-3> "global_selectFuncTag $t"
      }
    }
  }

  # -- do the same thing for class definitions
  foreach t $depfile(allctags) {
    set f [lindex $classes($t,file) 1]
    if { $f != "-" && [string compare $f $file] == 0 } {
      if {[scan $classes($t,pos) "%s %s %s %s" hstart hend bstart bend] == 4} {
        $win tag add class$t $hstart $bend
        $win tag configure class$t -background Gray95 -relief raised
        $win tag bind class$t <Button-1> "selectClassTag $t"
        $win tag bind class$t <Button-3> "global_selectClassTag $t"
      }
    }
  }
}

#
# readFile: read text of file into text window
#
#      win: pathname of text window
#     file: UNIX pathname of file to read
#

proc readFile {win file} {
  global depfile REMSH

  if { $depfile(host) == "localhost" } {
    set in [open $file r]
  } else {
    set in [open "|$REMSH $depfile(host) -n \"cd $depfile(dir); cat $file\"" r]
  }
  $win insert end [read $in]
  if [catch {close $in} errmsg] {
    showError "$file: `$errmsg'."
  }
  $win configure -state disabled
}

#
# showFile: show body of function or class in text window
#
#      tag: Sage++ id of selected item
#     what: type of selected item: Func, Class, or Line
#

proc showFile {tag what {fn ""} {ln -}} {
  global myself
  global depfile
  global funcs classes
  global showFile
  global TAUDIR

  # -- get necessary info from global database
  switch -exact $what {
    Func  {
            set file [lindex $funcs($tag,file) 1]
            set line [lindex $funcs($tag,file) 0]
            set name $funcs($tag,name)
          }
    Class {
            set file [lindex $classes($tag,file) 1]
            set line [lindex $classes($tag,file) 0]
            set name $classes($tag,name)
          }
    Line  {
            set file $fn
            set line $ln
            set name $funcs($tag,name)
          }
    default {
            puts stderr "$myself: internal error: invalid what $what"
            exit
          }
  }

  if { $file == "-" } {
    showError "Filename for function or class `$name' not specified."
    return
  } else {
    if { $depfile(host) == "localhost" && ! [file readable $file] } {
      showError "Cannot open file: `$file'."
      return
    }

    if { ! [winfo exists .viewer] } {
      # -- text window does not yet exist, create and display one
      toplevel .viewer
      wm title .viewer "$depfile(host):[normalizePath $depfile(dir)/$file]"
      wm iconbitmap .viewer @$TAUDIR/xbm/$myself.xbm
      wm minsize .viewer 50 50

      text .viewer.t1 -width 80 -height 32 -background white -foreground black
      scrollbar .viewer.s1 -orient vert -relief sunken \
                      -command ".viewer.t1 yview"
      .viewer.t1 configure -yscroll ".viewer.s1 set"
      button .viewer.bu -text "close" -command "destroy .viewer"

      pack .viewer.bu -side bottom -fill x
      pack .viewer.s1 -side right  -fill y
      pack .viewer.t1 -side top -expand yes -fill both

      # -- read necessay file, mark all functions and classes
      # -- highlight selected item
      readFile .viewer.t1 $file
      markAllFuncAndClass .viewer.t1 $file
      show$what .viewer.t1 $tag $file $line
    } else {
      # -- there is already a text window
      if { $file == $showFile(name) } {
        if { $tag != $showFile(tag) || $what != $showFile(what) } {
          # -- same file, but different selected item
          # -- un-highlight old on, highlight new one
          eval .viewer.t1 tag delete func body line \
                           [lpick [.viewer.t1 tag names] {child* base* ctype}]
          show$what .viewer.t1 $tag $file $line
        }
      } else {
        # -- different file; update title
        wm title .viewer "$depfile(host):[normalizePath $depfile(dir)/$file]"
        .viewer.t1 configure -state normal
        .viewer.t1 delete 1.0 end

        # -- un-highlight old selected items
        # -- this is necessary, otherwise tags which are used for highlightning
        # -- would be applied to new text
        eval .viewer.t1 tag delete body line \
             [lpick [.viewer.t1 tag names] {child* base* func* ctype class*}]

        # -- read new file, mark all functions and classes
        # -- highlight selected item
        readFile .viewer.t1 $file
        markAllFuncAndClass .viewer.t1 $file
        show$what .viewer.t1 $tag $file $line
      }
    }
    # -- remember currently selected item and its file
    set showFile(name) $file
    set showFile(tag)  $tag
    set showFile(what) $what
  }
}

set selectBox(File) "";    # currently selected file in Files listbox
set selectBox(Func) "";    # currently selected function in Functions listbox
set selectBox(Class) "";   # currently selected class in Classes listbox
set selectBox(Method) "";  # currently selected method in Methods listbox

#
# processFiles: process mouse buttons for Files listbox
#
#       button: number of pressed button
#          win: pathname of Files listbox
#            y: y coordinate within window where button was pressed
#

proc processFiles {button win y} {
  global myself
  global selectBox
  global funcs

  # -- ignore button 2 in File listbox
  if { $button == 2 } { return }

  # -- find item selected, bring it in the middle if necessary, and remember it
  set i [$win nearest $y]
  selectItem $win $i
  set selectBox(File) [.$myself.fi.l1 get $i]

  # -- update Function listbox according to select file
  set f $funcs($selectBox(File))
  redrawSelectBox .$myself.fu "Functions ($selectBox(File))" $f
}

#
# processFunctions: process mouse buttons for Functions listbox
#
#           button: number of pressed button
#              win: pathname of Functions listbox
#                y: y coordinate within window where button was pressed
#

proc processFunctions {button win y} {
  global selectBox
  global funcs

  # -- find item selected, and invoke global features depending on button
  set i [$win nearest $y]
  set t [lindex [lindex $funcs($selectBox(File)) $i] 1]

  if { $button == 2 } {
    global_showFuncTag $t
  } else {
    global_selectFuncTag $t
  }
}

#
# processClasses: process mouse buttons for Classes listbox
#
#         button: number of pressed button
#            win: pathname of Classes listbox
#              y: y coordinate within window where button was pressed
#

proc processClasses {button win y} {
  global myself
  global selectBox
  global allclasses classes

  # -- ignore button 2 in File listbox
  if { $button == 2 } { return }

  # -- find item selected, bring it in the middle if necessary, and remember it
  set i [$win nearest $y]
  selectItem $win $i
  set c [lindex $allclasses $i]
  set selectBox(Class) $c

  # -- update Methods listbox according to select class
  set m $classes($c,methods)
  redrawSelectBox .$myself.me "Methods ($selectBox(Class))" $m

  # -- invoke local / global selection depending on button
  if { $button == 1 } {
    selectClassTag $classes($c,tag) 0
  } else {
    global_selectClassTag $classes($c,tag)
  }
}

#
# processMethods: process mouse buttons for Methods listbox
#
#         button: number of pressed button
#            win: pathname of Methods listbox
#              y: y coordinate within window where button was pressed
#

proc processMethods {button win y} {
  global selectBox
  global classes

  # -- find item selected, and invoke global features depending on button
  set i [$win nearest $y]
  set t [lindex [lindex $classes($selectBox(Class),methods) $i] 1]

  if { $button == 2 } {
    global_showFuncTag $t
  } else {
    global_selectFuncTag $t
  }
}

#
# proc selectClassTag: implementation of global feature "selectClass"
#                      fancy selects class in Classes listbox and shows its
#                      members in the Members listbox
#                      Also, it highlights the body of the class declaration
#
#                ctag: Sage++ id of class to select
#            showdecl: highlight body?
#

proc selectClassTag {ctag {showdecl 1}} {
  global myself
  global classes allclasses
  global selectBox

  if [ info exists classes($ctag,name) ] {
    # -- class is known to fancy, select it
    set class $classes($ctag,name)
    if { $class != $selectBox(Class) } {
      # -- class not already selected, select it
      set i [lsearch -exact $allclasses $class]
      if { $i != -1 } { selectItem .$myself.cl.l1 $i }
      set selectBox(Class) $class

      # -- update Members listbox
      if [ info exists classes($class,methods) ] {
        set m $classes($class,methods)
      } else {
        set m {}
      }
      redrawSelectBox .$myself.me "Methods ($class)" $m
    }

    # -- highlight body, if necessary
    if { $showdecl } {
      showFile $ctag Class
    }
  } else {
    # -- class is unknown to fancy, un-select previous one
    set selectBox(Class) ""
    set selectBox(Method) ""
    redrawSelectBox .$myself.cl "Classes" $allclasses
    redrawSelectBox .$myself.me "Methods" {}
  }
}

#
# processTag: update listboxes if function is selected in text window
#
#        tag: Sage+= id of selected item
#

proc processTag {tag} {
  global myself
  global funcs classes
  global selectBox
  global allfiles allclasses

  set class [lindex $funcs($tag,class) 0]
  set name     $funcs($tag,name)

  if { $class == "-" } {
    # -- selected function is NOT member function: get file
    set file [file tail [lindex $funcs($tag,file) 1]]
    if { $file == "-" } { return }
    set list $funcs($file)

    # -- update Files listbox, if necessary
    if { $file != $selectBox(File) } {
      set i [lsearch -exact $allfiles $file]
      selectItem .$myself.fi.l1 $i
      set selectBox(File) $file
      redrawSelectBox .$myself.fu "Functions ($file)" $list
    }

    # -- update Functions listbox
    set i [lsearch -exact $list [list $name $tag]]
    selectItem .$myself.fu.l1 $i
    set selectBox(Func) $tag
  } else {
    # -- selected function is member function: get class
    set list $classes($class,methods)

    # -- update Classes listbox, if necessary
    if { $class != $selectBox(Class) } {
      set i [lsearch -exact $allclasses $class]
      selectItem .$myself.cl.l1 $i
      set selectBox(Class) $class
      redrawSelectBox .$myself.me "Methods ($class)" $list
      selectClassTag $classes($class,tag) 0
    }

    # -- update methods listbox
    set i [lsearch -exact $list [list $name $tag]]
    selectItem .$myself.me.l1 $i
    set selectBox(method) $tag
  }
}

#
# selectItem: select and display item in listbox
#
#        win: pathname of listbox
#      index: index of selected item
#

proc selectItem {win index} {
  # -- select item
  $win select from $index

  # -- scroll listbox to bring item on screen, if necessary
  set top [$win nearest 0]
  if { $index > [expr $top+9] || $index < $top } {
    $win yview [expr $index-5]
  }
}

#
# selectBox: create and display new selectbox (listbox+title+scrollbar)
#
#       win: pathname of selectBox
#     title: title for selectBox
#     names: list of items to display in listbox
#

proc selectBox {win title names} {
  frame $win
  label $win.t -relief raised -text $title
  listbox $win.l1 -relief sunken -geometry 30x10
  tk_listboxSingleSelect $win.l1
  foreach n $names {
    $win.l1 insert end [lindex $n 0]
  }
  scrollbar $win.s1 -orient vert -relief sunken -command "$win.l1 yview"
  $win.l1 configure -yscroll "$win.s1 set"
  scrollbar $win.s2 -orient horiz -relief sunken -command "$win.l1 xview"
  $win.l1 configure -xscroll "$win.s2 set"

  pack $win.t -side top -fill x
  pack $win.s1 -side right -fill y
  pack $win.l1 $win.s2 -side top -fill x

  bind $win.l1 <Button-1> "process$title 1 $win.l1 %y"
  bind $win.l1 <Button-2> "process$title 2 $win.l1 %y"
  bind $win.l1 <Button-3> "process$title 3 $win.l1 %y"
  return $win
}

#
# redrawSelectBox: update selectBox
#
#             win: pathname of selectBox
#        newtitle: new title for selectBox
#           names: new list of items to display in listbox
#

proc redrawSelectBox {win newtitle names} {
  $win.l1 delete 0 end
  foreach n $names {
    $win.l1 insert end [lindex $n 0]
  }
  $win.t configure -text $newtitle
}

#
# loadDep: switch to another application
#          this is invoked from the TAU master control window
#
#    path: UNIX pathname of depfile to read
#    root: patchname of SAGEROOT
#    arch: architecture of machine where depfile resides
#

proc loadDep {path root arch} {
  global myself
  global funcs depfile
  global classes allclasses allfiles
  global selectBox

  if { ! [regexp {(.*):(.*)} $path dummy depfile(host) path] } {
    set depfile(host) "localhost"
  }
  set depfile(root) $root
  set depfile(arch) $arch

  # -- delete old database of functions and classes
  unset funcs classes allclasses allfiles

  # -- read new depfile and rebuild database
  if { [readDep -dumptxt $path] == "NOT_OK" } {
    exit
  }
  computeAddInfo

  # -- get rid of old viewer
  if [ winfo exists .viewer ] { destroy .viewer }

  # -- reset litboxes
  set selectBox(File) ""
  set selectBox(Func) ""
  set selectBox(Class) ""
  set selectBox(Method) ""
  redrawSelectBox .$myself.fi "Files" $allfiles
  redrawSelectBox .$myself.cl "Classes" $allclasses
  redrawSelectBox .$myself.fu "Functions" {}
  redrawSelectBox .$myself.me "Methods" {}
}

#
# computeAddInfo: setup additional information
#                 after reading a new depfile with loadDep
#           

proc computeAddInfo {} {
  global depfile
  global funcs
  global classes allclasses allfiles

  foreach t $depfile(allftags) {
    set classtag  [lindex $funcs($t,class) 2]
    set classname [lindex $funcs($t,class) 0]
    set file [file tail [lindex $funcs($t,file) 1]]

    if { $file != "-" } {
      # -- function has body
      # -- setup list of all files used (allfiles)
      if { ! [info exists funcs($file)] } {
        lappend allfiles $file
        set funcs($file) ""
      }

      if { $classtag == -2 } {
        # -- NOT a member function, add function tag to file specific list
        lappend funcs($file) [list $funcs($t,name) $t]
      } else {
        # -- member function, setup list of all classes and their members
        if { ! [info exists classes($classname,tag)] } {
          lappend allclasses $classname
          set classes($classname,tag) $classtag
        }
        lappend classes($classname,methods) [list $funcs($t,name) $t]
      }
    }
  }

  # -- sort generated lists alphabetically
  set allclasses [lsort $allclasses]
  set allfiles [lsort $allfiles]
  foreach f $allfiles {
    set funcs($f) [lsort $funcs($f)]
  }
  foreach c $allclasses {
    set classes($c,methods) [lsort $classes($c,methods)]
  }
}

#
# createWindow: create and display main window of fancy
#

proc createWindow {} {
  global TAUDIR
  global myself
  global allclasses allfiles

  toplevel .$myself
  wm title .$myself "FANCY"
  wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm

  frame .$myself.mbar -relief raised -borderwidth 2
  menubutton .$myself.mbar.b1 -text File -menu .$myself.mbar.b1.m1 -underline 0
  menu .$myself.mbar.b1.m1
  .$myself.mbar.b1.m1 add command -label "Exit" -underline 0 -command "exit"

  menubutton .$myself.mbar.b2 -text Help -menu .$myself.mbar.b2.m1 -underline 0
  menu .$myself.mbar.b2.m1
  .$myself.mbar.b2.m1 add command -label "on $myself" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1-$myself 1\]"
  .$myself.mbar.b2.m1 add command -label "on viewer" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.1-viewer 1\]"
  .$myself.mbar.b2.m1 add separator
  .$myself.mbar.b2.m1 add command -label "on using help" -underline 3 \
                   -command "xsend tau {showHelp general 1-help 1}"

  pack .$myself.mbar.b1 -side left -padx 5
  pack .$myself.mbar.b2 -side right -padx 5
  pack .$myself.mbar -side top -fill x
  tk_menuBar .$myself.mbar .$myself.mbar.b1 .$myself.mbar.b2

  frame .$myself.left
  selectBox .$myself.fi Files $allfiles
  selectBox .$myself.cl Classes $allclasses
  pack .$myself.fi .$myself.cl -side top -padx 15 -pady 15 -in .$myself.left
  pack .$myself.left -side left

  frame .$myself.right
  selectBox .$myself.fu Functions {}
  selectBox .$myself.me Methods {}
  pack .$myself.fu .$myself.me -side top -padx 15 -pady 15 -in .$myself.right
  pack .$myself.right -side right
  focus .$myself.mbar
}

# ------------
# -- main code
# ------------
if {$argc == 2 || $argc > 3} {
  puts stderr "usage: $myself \[\[host:\]depFile \| \[host:\]directory\]"
  exit
}

if {$argc == 1} {
  set dfile [findDep [lindex $argv 0]]
} elseif {$argc == 3} {
  set depfile(root) [lindex $argv 1]
  set depfile(arch) [lindex $argv 2]
  set dfile [findDep [lindex $argv 0]]
} else {
  set dfile [findDep .]
}

if { $dfile == "NOT_OK" || [readDep -dumptxt $dfile] == "NOT_OK" } {
  exit
}

computeAddInfo
createWindow
launchTAU
