/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/***************************************************************************
 *                                                                         *
 *                                                                         *
 *                HPC++ Routine for managing access to data                *
 *                Sequential Set Of routine                                *
 *                                                                         *
 *                F. Bodin, Indiana University                             *
 *                                                                         *
 ***************************************************************************/
#include <stdarg.h>

#define MaxCollection 100

#include <stdio.h>
#include "GlobalStruct.h"

#ifndef __ksr__
#  define __private
#endif

__private static int NumberOfCollection = 0;
__private static PTR_INFO TableOfCollection[MaxCollection];
__private static char *CurrentCollectionTraversal[MAXPARAL];
__private static int TraversalLevel = -1;
__private static int WarningCount =0;

/* Forward definition of activated function */
void GiveWhereDataIs();
void SendData();
void UpdateMyData();

/* Forward definition of local management */
int WhereIsData();
int AskForData();
void Message();

/**************************************************************************
 ********************** General Routine ***********************************
 **************************************************************************/
#include "common.h" /* Common support routines for EVERY KernelCom.c (phb) */

/*************************************************************/
void Message(st, flag)
char *st;
int flag;
{
  if (flag == WARNING)
    {
      WarningCount++;
      fprintf(stderr, "Warning: %s\n",st);
      return;
    }
  if (flag == FATAL)
    {
      fprintf(stderr, "Fatal error: %s\n",st);
      return;
    }
}

int GetNumberOfCollection()
{
  return NumberOfCollection;
}

void IncNumberOfCollection()
{
  if (NumberOfCollection >= MaxCollection)
    Message("Too many collections.",FATAL);
  NumberOfCollection++;
}

void SetTableOfCollection (num, addr)
int num;
char *addr;
{
  TableOfCollection[num] = (PTR_INFO) addr;

}

void SetNextCollectionTraversalLevel (addr)
char *addr;
{
  TraversalLevel++;
  if (TraversalLevel >= MAXPARAL)
    Message("ResetLocal(): Too much parallelism.",FATAL);
  CurrentCollectionTraversal[TraversalLevel] = addr;
}

char *GetCollectionTraversalLevel ()
{
  return CurrentCollectionTraversal[TraversalLevel];
}

void DecCollectionTraversalLevel ()
{
  TraversalLevel--;
}

void ResetCollectionTraversalLevel ()
{
  TraversalLevel = -1;
}

/**************************************************************************
 ********************** Kernel Request ************************************
 **************************************************************************/

int IamtheManager(id)
int id;
{
  Message("IamtheManager called in shared memory mode",FATAL);
  return 1;
}

int WhereIsData(id,info)
int id;
int info;
{
  static int manager = 0;
  return 0;
}

int AskForDataBlock(int where, int index, int offset, 
                       int Coll_Id, int size, int *buffer)
{
  char *p;
  p = (char*) (TableOfCollection[Coll_Id]->Table[index]+offset);
  memcpy(buffer,p,size);
  return 0;
}

main(argc, argv)
int argc;
char **argv;
{
  pcxx_InitParam (&argc, argv);
  SMRTSSetup ();
  SMRTSFork (argc, argv);
}

void pcxx_MainBegin(int *argc, char *argv[]) { }
void pcxx_MainEnd() {}

extern int  MyTotalNodes;  /* NO __private!! */
__private extern int  MyNodeNumber;

int pcxx_TotalNodes() { return(MyTotalNodes); }

int pcxx_NodeNumber() { return(MyNodeNumber); }

extern void RemoteFunctionList(int *element_address,int function_number,
                               char *arg_buffer);

void AskForRemoteExecute(int CollectionNumber,int element_index,
                         int function_number,char *arg_buffer,int size,
                         long tag_location)
{
  /* this function needs to be modified to work properly on the shared memory.
     somehow, a mechanism has to be devised to lock a memory location
     when it is being modified by a certain processor.  right now this
     function works if it is guaranteed that the memory location is modified
     sequentially
  */
  long *p;
  RemoteFunctionList(TableOfCollection[CollectionNumber]->Table[element_index],
                     function_number, (char*) arg_buffer);
  if(tag_location != 0) {
    p = (long*)tag_location;
    (*p)++;
  }
}

int AskForFArray(int where, unsigned index, unsigned Coll_Id, int offset,
                       char* buffer, int size)
{
  char *pt;
  pt = (char*)*(TableOfCollection[Coll_Id]->Table[index]+offset);
  memcpy(buffer,pt,size);
  return 0;
}

void CopyFArray(char *dest,char *source,int rank,int size)
{
  memcpy(dest,source,size);
}

void CopyFArrayElem(char* source, char *buffer,
                  int elem_size, int rank,
                  int i_len, int i_inc, int i_off, 
                  int j_len, int j_inc, int j_off,
                  int k_len, int k_inc, int k_off, 
                  int i, int j, int k)
{
  int m;
  char *pt;
  pt = source + (i + j*i_len + k*i_len*j_len)*elem_size;
  for(m = 0; m < elem_size; m++) buffer[m] = pt[m];
}

int AskForFArrayElem(int where,int index,int Coll_Id,int offset,char *buffer,
                   int elem_size, int rank,
                   int i_len, int i_inc, int i_off, 
                   int j_len, int j_inc, int j_off,
                   int k_len, int k_inc, int k_off, 
                   int i, int j, int k)
{
  char *pt;
  pt = (char*)*(TableOfCollection[Coll_Id]->Table[index]+offset);
  pt = pt + (i + j*i_len + k*i_len*j_len)*elem_size;
  memcpy(buffer,pt,elem_size);
  return 0;
}

/* the following reduction function have to make use of the machine native
   functions */

void pcxx_ReduceIntAdd(void* buffer,int elem_count) {}

void pcxx_ReduceFloatAdd(void* buffer,int elem_count) {}

void pcxx_ReduceDoubleAdd(void* buffer,int elem_count) {}

void pcxx_ReduceDoubleMax(void* buffer,int elem_count) {}

void pcxx_ReduceFloatMax(void* buffer,int elem_count) {}

void pcxx_ReduceIntMax(void* buffer,int elem_count) {}

void pcxx_ReduceDoubleMin(void* buffer,int elem_count) {}

void pcxx_ReduceFloatMin(void* buffer,int elem_count) {}

void pcxx_ReduceIntMin(void* buffer,int elem_count) {}

