/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# ifndef PROF_INITIALIZED
#   define PROF_INITIALIZED
#   include "profiler_defs.h"

#   ifdef PROFILING_ON

P_COLL_DESCR *p_coll_list[MaxCollection];
P_FUNC_DESCR *p_func_list;
char *p_collname, *p_elemname;
P_CQUEUE *p_varname = 0;

int p_curr_func = -1;

class Profiler {
    int parent;
    int ident;
    int numfunc;
    double start;
public:
    Profiler (int id);
    Profiler (int id, char countonly);
    Profiler (int id, int num);
    ~Profiler ();
};

Profiler :: Profiler (int id)
{
  parent  = p_curr_func;
  p_curr_func = ident = id;
  numfunc = 0;
  start   = pcxx_GetUSecD();
}

Profiler :: Profiler (int id, char countonly)
{
  parent  = p_curr_func;
  p_curr_func = ident = id;
  numfunc = -1;
}

Profiler :: Profiler (int id, int num)
{
  p_func_list = new P_FUNC_DESCR[num];
  for (int i=0; i<num; i++) {
      p_func_list[i].numcalls = 0;
      p_func_list[i].usec     = 0.0;
      p_func_list[i].cumusec  = 0.0;
  }
  parent  = -1;
  p_curr_func = ident = id;
  numfunc = num;
  start   = pcxx_GetUSecD();

  for (i=0; i<MaxCollection; i++) p_coll_list[i] = 0;
}

Profiler :: ~Profiler ()
{
  if ( numfunc < 0 ) {
    // count only destructor
    p_func_list[ident].numcalls++;
    p_curr_func = parent;
  }
  else {
    double duration = pcxx_GetUSecD() - start;
    p_func_list[ident].numcalls++;
    p_func_list[ident].usec    += duration;
    p_func_list[ident].cumusec += duration;

    if ( parent == -1 ) {
        // last exit: dump results
        int numcoll, n;
        int out;
        int *buf, *ptr;
        char *profname;

        profname = new char[12];
        sprintf (profname, "profile.%d", pcxx_NodeNumber());
        if ( (out = creat (profname, 0644)) == -1 ) {
            perror (profname);
            exit (1);
        }
        numcoll = GetNumberOfCollection();

        ptr = buf = new int[3+numfunc*5+numcoll*5];
        *ptr++ = 0x00010203;
        *ptr++ = numfunc;
        *ptr++ = numcoll;

        for (int i = 0; i<numfunc; i++) {
          *ptr++ = p_func_list[i].numcalls;
          if (p_func_list[i].usec > 0.0 ) {
            *ptr++ = n = (int) (p_func_list[i].usec / 1e6);
            *ptr++ = (int) (p_func_list[i].usec - n*1e6);
            *ptr++ = n = (int) (p_func_list[i].cumusec / 1e6);
            *ptr++ = (int) (p_func_list[i].cumusec - n*1e6);
          }
          else {
            *ptr++ = 0; *ptr++ = 0; *ptr++ = 0; *ptr++ = 0;
          }
        }

        for (i = 0; i<numcoll; i++) {
          *ptr++ = p_coll_list[i]->numelem;
          *ptr++ = p_coll_list[i]->dim;
          *ptr++ = p_coll_list[i]->size;
          *ptr++ = p_coll_list[i]->localacs;
          *ptr++ = p_coll_list[i]->remoteacs;
        }
        write (out, (void *) buf, (3+numfunc*5+numcoll*5) * sizeof(int));
        delete [] buf;

        for (i = 0; i<numcoll; i++) {
          write (out, (void *) (p_coll_list[i]->collname),
                                            strlen(p_coll_list[i]->collname)+1);
          write (out, (void *) (p_coll_list[i]->elemname),
                                            strlen(p_coll_list[i]->elemname)+1);
          write (out, (void *) (p_coll_list[i]->varname),
                                            strlen(p_coll_list[i]->varname)+1);
        }
        close (out);
    }
    else {
        // child: restore current func
        p_curr_func = parent;

        // substract child time form parent's time
        p_func_list[parent].usec -= duration;
    }
  }
}

void PROFILE_COLL_NAMES (char *coll, char *elem, char *var) {
  P_CQUEUE *v;

  p_collname = coll;
  p_elemname = elem;

  if ( v = p_varname ) {
    while ( v->next ) v = v->next;
    v = v->next = new P_CQUEUE;
  }
  else {
    v = p_varname = new P_CQUEUE;
  }
  v->name = var;
  v->next = 0;
}
#   endif
# endif
