/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# ifndef PROFILE_ENTER
#   ifdef PROFILING_ON

#     ifdef __cplusplus

extern "C" int creat(char *, int);
extern "C" int close(int);
extern "C" int write(int, void *, int);
extern "C" void perror(const char *);
extern "C" void exit(int);
#ifndef __GNUC__
extern "C" int sprintf(char * , const char * ,  ...);
# ifdef __ksr__
extern "C" unsigned long strlen(const char *s);
# else
extern "C" unsigned strlen(const char *s);
# endif
# endif

#     endif

#     ifndef MaxCollection
#       define MaxCollection 50
#     endif

typedef struct p_func_descr {
    int     numcalls;
    double  usec;
    double  cumusec;
} P_FUNC_DESCR;

typedef struct p_coll_descr {
    int numelem;
    int dim;
    int size;
    int localacs;
    int remoteacs;
    char *collname;
    char *elemname;
    char *varname;
} P_COLL_DESCR;

typedef struct p_cqueue {
    char *name;
    struct p_cqueue *next;
} P_CQUEUE;

#     ifndef PROF_INITIALIZED
#       if defined(__ksr__) && ! defined(UNIPROC)
          extern __private P_COLL_DESCR *p_coll_list[MaxCollection];
          extern __private P_FUNC_DESCR *p_func_list;
          extern __private int p_curr_func;
          extern __private char *p_collname, *p_elemname;
          extern __private P_CQUEUE *p_varname;
#       else
          extern P_COLL_DESCR *p_coll_list[MaxCollection];
          extern P_FUNC_DESCR *p_func_list;
          extern int p_curr_func;
          extern char *p_collname, *p_elemname;
          extern P_CQUEUE *p_varname;
#       endif
#     endif

#     define PROFILE_ENTER(id)           \
        int prf_parent = p_curr_func;    \
        double prf_start = pcxx_GetUSecD(); \
        p_curr_func = id;

#     define PROFILE_EXIT(id)                         \
        do {                                          \
          double duration = pcxx_GetUSecD() - prf_start; \
          p_func_list[id].numcalls++;                 \
          p_func_list[id].usec    += duration;        \
          p_func_list[id].cumusec += duration;        \
          p_curr_func = prf_parent;                   \
          p_func_list[prf_parent].usec -= duration;   \
        } while (0)

#     define PROFILE_COLL_INIT(n,d,s)                                        \
        do {                                                                 \
          p_coll_list[CollectionNumber] = (P_COLL_DESCR *) new P_COLL_DESCR; \
          p_coll_list[CollectionNumber]->numelem = n;                        \
          p_coll_list[CollectionNumber]->dim     = d;                        \
          p_coll_list[CollectionNumber]->size    = s;                        \
          p_coll_list[CollectionNumber]->localacs  = 0;                      \
          p_coll_list[CollectionNumber]->remoteacs = 0;                      \
          p_coll_list[CollectionNumber]->collname  = p_collname;             \
          p_coll_list[CollectionNumber]->elemname  = p_elemname;             \
          if ( p_varname->name ) {                                           \
            p_coll_list[CollectionNumber]->varname = p_varname->name;        \
            p_varname = p_varname->next;                                     \
          } else                                                             \
            p_coll_list[CollectionNumber]->varname = "???";                  \
        } while (0)

#     define PROFILE_LOCAL_ACS() \
        p_coll_list[CollectionNumber]->localacs++

#     define PROFILE_REMOTE_ACS() \
        p_coll_list[CollectionNumber]->remoteacs++

#     define PROFILE_Barrier 1
#     define PROFILE_Poll    2
#     define PROFILE_Where   3

#   else
#     define PROFILE_COLL_INIT(n,d,s)
#     define PROFILE_LOCAL_ACS()
#     define PROFILE_REMOTE_ACS()
#     define PROFILE_ENTER(id)
#     define PROFILE_EXIT(id)
#     define PROFILE_COLL_NAMES(c,e,v)
#   endif
# endif
