/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* Written by Pete Beckman (by modifying Bernd Mohr's symmetry version) */
/* 6/22/94 */

/*==============================================================*/
/* sgilib : SGI run-time support library			*/
/*==============================================================*/

#include <sys/types.h>
#include <sys/prctl.h>
#include <ulocks.h>
#include <task.h>
#include <stdio.h>
#include <sys/sysmp.h>
#include <unistd.h>
 
#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"
#include "brkExecInc.h"

extern pcxx_PARAM_t pcxx_param;

extern void New_Main();

#define BCASTMAX 1024

/*******************************************************************/
/* Shared data structures */
/*******************************************************************/
barrier_t *SGI_Gen_Barr;       /* general-purpose barrier */
barrier_t *SGI_Done_Barr;      /* Done with everything barrier */
void **SGI_Elem_Alloc_Vector;  /* vector of elem space pointers */
char *BroadcastBuff;          /* a place to store broacast values */

int *SharedHandle;      /* Name of "Arena" created by usinit() */
int MyTotalNodes = -1;     /* number of processors */
int MyNodeNumber = -1;     /* who I am */

/*
 @ SMRTSSetup() : set up to run on the SGI
 */

int SMRTSSetup()
{
  char fname[1024];

  /* allocate processors */
  if (pcxx_param.numproc >= sysmp(MP_NAPROCS)){
    MyTotalNodes = sysmp(MP_NAPROCS);
  } else {
    MyTotalNodes = pcxx_param.numproc;
  }

  MyNodeNumber = MyTotalNodes - 1;
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);

  usconfig(CONF_INITUSERS,20);              /* Up to 20 processors */
  usconfig(CONF_INITSIZE,1024*1024*128);    /* 128 Meg of shared data */
  usconfig(CONF_ARENATYPE,US_SHAREDONLY);   /* ulink memmap after open */

  /* Open unique memmap filename */
  sprintf(fname,"/usr/tmp/pC++%d.memmap",getpid());
  SharedHandle = usinit(fname);

  /* initialize barriers */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);
  SGI_Gen_Barr = new_barrier(SharedHandle);
  init_barrier(SGI_Gen_Barr);
  SGI_Done_Barr = new_barrier(SharedHandle);
  init_barrier(SGI_Done_Barr);

  /* initialize shared alloc location */
  SGI_Elem_Alloc_Vector = 
    (void **) usmalloc (MyTotalNodes * sizeof(void *),SharedHandle);

  /* allocate Shared Broadcast Buffer */
  BroadcastBuff = (char *) usmalloc(BCASTMAX,SharedHandle);

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return(0);
}

/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork(argc, argv)
  int argc;
  char **argv;
{
  int k;
  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

  for (k=0; k<MyTotalNodes-1; k++) 
    sproc((void *) New_Main,0,k);

/*    fprintf(stderr, "m_fork failed\n");
    exit(1);*/

  /* Go do New_Main() myself! */
  New_Main((void *) (MyTotalNodes-1),1024);
  barrier(SGI_Done_Barr,MyTotalNodes);

  /* finished */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  return(0);
}

/*
 @ SMRTSNewProc : more book-keeping for newly-created process
 */

int SMRTSNewProc(num)
  int num;
{
  /* record my ID */
  MyNodeNumber = num;
  pcxx_InitClocks ();
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  return(MyNodeNumber);
}

/*
 @ barrier : wait at general barrier
 */

void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  brk_BREAKPOINT_EXECUTIVE();
  barrier(SGI_Gen_Barr,MyTotalNodes);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*
 @ pcxx_BrodcastBytes : sends messages from 0 to all other nodes
 */

void pcxx_BroadcastBytes(source, length, buffer)
int source;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  if ( length > BCASTMAX ) Message("Broadcast Too BIG!! ",FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (source == 0) for (i=0; i<length; i++) c[i] = p[i];
  pcxx_Barrier();
  if (source != 0) for (i=0; i<length; i++) p[i] = c[i];
}

/*
 @ SMRTSAllocElemSpace : allocate shared memory for proc in global vector
 */

void SMRTSAllocElemSpace (proc, size)
  int proc;
  int size;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_ALLOC_ELEM_ENTER, proc);
  SGI_Elem_Alloc_Vector[proc] = (void *) usmalloc (size,SharedHandle);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_ALLOC_ELEM_EXIT, (long) SGI_Elem_Alloc_Vector[proc]);
}

/*
 @ SMRTSGetElemSpacePtr : return proc entry of global vector
 */

void *SMRTSGetElemSpacePtr (proc)
  int proc;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_GET_ELEM_PTR, proc);
  return (SGI_Elem_Alloc_Vector[proc]);
}

/*
 @ SMRTSDealloc : deallocate shared memory pointers
 */

void SMRTSDealloc(ptr)
void *ptr;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  usfree(ptr,SharedHandle);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}

void SMRTSWaitToEnd() {
  if ((int) MyNodeNumber != (MyTotalNodes-1)) barrier(SGI_Done_Barr,MyTotalNodes);
}
