/***************************************************************************
 * Program:         ENVOY EUI Communication Library
 *
 * Module:          mpctof.c
 *
 * Purpose:         EUI  mpc (c) to mp (fortran c) interface routines
 *
 * Classification:  IBM Internal Use Only
 *
 * Author:          Peter de Jong
 *                  HPSSL
 *
 * Modified:        David Edelsohn
 *                  IBM T.J. Watson Research Center
 *
 * Fri Apr 23 14:27:28 1993  David Edelsohn  (c1dje@tweetie)
 *
 *      * remove mpc_init() and mpc_end().
 *
 *	* implement mpc_pack() and mpc_unpack() directly based upon
 *        _ccl_pack0 and _ccl_unpack0 in EPACK.C from Alex W. Ho
 *        (Almaden FMPC) instead of calling _ccl routines.
 *
 *      * pass return values
 *
 ***************************************************************************/

#include "mpceui.h"
#include "mpeui.h"

int mpc_environ (int *numtask, int *taskid)
{
    return mp_environ (numtask, taskid);
}

int mpc_stopall (int errcode)
{
    return mp_stopall (&errcode);
}

int mpc_task_query (int nbuf[], int nelem, int qtype)
{
    return mp_task_query (nbuf, &nelem, &qtype);
}

int mpc_task_set (int nbuf[], int stype)
{
    return mp_task_set (nbuf, &stype);
}

int mpc_brecv (char *inmsg, int msglen, int *source, int *type, int *nbytes)
{
    return mp_brecv (inmsg, &msglen, source, type, nbytes);
}

int mpc_bsend (char *outmsg, int msglen, int dest, int type)
{
    return mp_bsend (outmsg, &msglen, &dest, &type);
}

int mpc_bsendrecv (char *outmsg, int outmsglen, int dest, int type,
		   char *inmsg, int inmsglen, int *source, int *nbytes)
{
    return mp_bsendrecv (outmsg, &outmsglen, &dest, &type,
			 inmsg, &inmsglen, source, nbytes);
}

/*
int mpc_bvrecv (char *inmsg, int blklen, int offset, int blknum,
		int *source, int *type, int *nbytes)
{
    return mp_bvrecv (inmsg, &blklen, &offset, &blknum,
		      source, type, nbytes);
}

int mpc_bvsend (char *outmsg, int blklen, int offset, int blknum,
		int dest, int type)
{
    return mp_bvsend (outmsg, &blklen, &offset, &blknum, &dest, &type);
}
*/

int mpc_recv (char *inmsg, int msglen, int *source, int *type, int *msgid)
{
    return mp_recv (inmsg, &msglen, source, type, msgid);
}

int mpc_send (char *outmsg, int msglen, int dest, int type, int *msgid)
{
    return mp_send (outmsg, &msglen, &dest, &type, msgid);
}

int mpc_status (int msgid)
{
    return mp_status (&msgid);
}

/*
int mpc_vrecv (char *inmsg, int blklen, int offset, int blknum,
	       int *source, int *type, int *msgid)
{
    return mp_vrecv (inmsg, &blklen, &offset, &blknum,
		     source, type, msgid);
}

int mpc_vsend (char *outmsg, int  blklen, int offset, int blknum,
	       int dest, int type, int *msgid)
{
    return mp_vsend (outmsg, &blklen, &offset, &blknum, &dest, &type, msgid);
}
*/

int mpc_wait (int *msgid, int *nbytes)
{
    return mp_wait (msgid, nbytes);
}


int mpc_bcast (char *msg, int msglen, int source, int gid)
{
    return mp_bcast(msg, &msglen, &source, &gid);
}


int mpc_combine (char *outmsg, char *inmsg, int msglen, int func(), int gid)
{
    return mp_combine (outmsg, inmsg, &msglen, func, &gid);
}

int mpc_concat (char *outmsg, char *inmsg, int blklen, int gid)
{
    return mp_concat (outmsg, inmsg, &blklen, &gid);
}

int mpc_gather (char *outmsg, char *inmsg, int blklen, int dest, int gid)
{
    return mp_gather (outmsg, inmsg, &blklen, &dest, &gid);
}


int mpc_index (char *outmsg, char *inmsg, int blklen, int gid)
{
    return mp_index (outmsg, inmsg, &blklen, &gid);
}

int mpc_prefix (char *outmsg, char* inmsg, int msglen, int func(), int gid)
{
    return mp_prefix (outmsg, inmsg, &msglen, func, &gid);
}

int mpc_reduce (char *outmsg, char *inmsg, int msglen, int dest,
		int func(), int gid)
{
    return mp_reduce (outmsg, inmsg, &msglen, &dest, func, &gid);
}

int mpc_scatter (char *outmsg, char *inmsg, int blklen, int source, int gid)
{
    return mp_scatter (outmsg, inmsg, &blklen, &source, &gid);
}

int mpc_shift (char *outmsg, char* inmsg, int msglen, int step,
	       int flag, int gid)
{
    return mp_shift (outmsg, inmsg, &msglen, &step, &flag, &gid);
}

int mpc_sync (int gid)
{
    return mp_sync(&gid);
}

int mpc_getlabel (int *label, int gid)
{
    return mp_getlabel (label, &gid);
}

int mpc_getmembers (int glist[], int gid)
{
    return mp_getmembers (glist, &gid);
}

int mpc_getrank (int *rank, int taskid, int gid)
{
    return mp_getrank (rank, &taskid, &gid);
}

int mpc_getsize (int *gsize, int gid)
{
    return mp_getsize (gsize, &gid);
}

int mpc_gettaskid  (int rank, int *taskid, int gid)
{
    return mp_gettaskid (&rank, taskid, &gid);
}

int mpc_group (int gsize, int glist[], int label, int *gid)
{
    return mp_group(&gsize, glist, &label, gid);
}

int mpc_partition (int parent_gid, int key, int label, int *gid)
{
    return mp_partition (&parent_gid, &key, &label, gid);
}

int mpc_pack (char *buf, char *msg, int lenblk, int stride, int nblk)
{
    int i, j, j1, j2;

    j1=0; j2=0;
    for (j=0; j<nblk; j++) {
	for (i=0; i<lenblk; i++)
	    msg[j1+i] = buf[j2+i];
	j1 += lenblk;
	j2 += stride;
    }
    return 0;
}

int mpc_unpack (char *msg, char *buf, int lenblk, int stride, int nblk)
{
    int i, j, j1, j2;

    j1=0; j2=0;
    for (j=0; j<nblk; j++) {
	for (i=0; i<lenblk; i++)
	    buf[j2+i] = msg[j1+i];
	j1 += lenblk;
	j2 += stride;
    }
    return 0;
}

