/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/***************************************************************************
 *                pC++ Routine for managing access to data                 *
 *                IBM SP1 Version                                          *
 *                                                                         *
 *                                                                         *
 *                Written by P. Beckman                                    *
 *                Based on Paragon version by D. Gannon                    *
 *                Based on cm5 version by F. Bodin                         *
 *                Modified by S. Yang (2/11/94)                            *
 *                Reworked by Pete Beckman (2/19/94)                       *
 ***************************************************************************/
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

#include "passive.h"

extern pcxx_PARAM_t pcxx_param;

/**************************************************************************/
/**************************************************************************/
/* MACHINE DEPENDANT CODE */
/**************************************************************************/
/**************************************************************************/
#include "mpctof.c"
#define PENDING  -1   /* Message buffer still busy */
#define INACTIVE -2   /* This may be bad, but IBM docs very vague */
                      /* What exactly is an "inactive" message? */
static int RecvMsgId, SendMsgId;
static int MsgType=DONTCARE, MsgFromWhom=DONTCARE;
extern int pcxx_MsgBlockSize;
/**************************************************************************/
/**************************************************************************/
/* Machine Dependant Message Prototypes:
/* IBM SP1 */
/**************************************************************************/
/**************************************************************************/
/* -- This section commented out since the header file declares them  */
/*
extern int mpc_recv(char *buff, int len, int *src, int *type, int *msgid);
extern int mpc_send(char *buff, int len, int dest, int type, int *msgid);
extern int mpc_bsend(char *buff, int len, int dest, int type);
extern int mpc_environ(int *numtasks, int *taskid);
extern int mpc_status(int msgid);
extern int mpc_stopall(int returnval);
*/

/*******************************************************************/
void pcxx_Abort()
  { mpc_stopall(1); exit(1); }

/*******************************************************************/
/* Post a recv for later messages */
void MyRecv()
{
  int rc;
  
  MsgFromWhom = DONTCARE;  MsgType=DONTCARE;  RecvMsgId = -1;

  /* Post a receive for ANY messages, they will arrive into ReceiveBuffer */
  rc = mpc_recv((char *) ReceiveBuffer, MAXBUFSIZE, 
                &MsgFromWhom, &MsgType, &RecvMsgId);
  if (rc < EUI_SUCCEED) Message("MyRecv: mpc_recv failed",FATAL);

  if (VERBOSE)
    printf("%d posted a recv: buffer:%x size:%d from:%d type:%d mid:%d)\n",
    MyNodeNumber,ReceiveBuffer,MAXBUFSIZE,MsgFromWhom,MsgType,RecvMsgId);
}

/*******************************************************************/
/* Send a buffer to a destination */
void MySend(int type, char * buffer, int len, int ToWhom)
{
  int rc,nb,x=0;

  if (len > MAXBUFSIZE) {
    fprintf(stderr,"pC++_ERROR: Buffer length %d greater than MAXBUFSIZE\n",
	    len,MAXBUFSIZE);
    Message("MySend: length > MAXBUFSIZE",FATAL);
  }

  if (SendMsgId != -1) {
    /* This is _NOT_ the first send */

    /* Wait for previous send to complete */
    /* rc = mpc_wait(&SendMsgId,&nb); */ 

    /* Did the previous send complete? */
    while (mpc_status(SendMsgId) == PENDING) {
      pcxx_Poll();    /* I still have to answer messages */
      printf(stderr,"%d: Waiting for Send Buffer to empty...\n",
	     MyNodeNumber);
    }
  }
  else {
    if (VERBOSE) printf("%d: Sending first message\n",MyNodeNumber);
  }

  /* Actually send the buffer (non-blocking) */
  rc = mpc_send(buffer, len, ToWhom, type, &SendMsgId);
  if (rc != 0) Message("MySend: mpc_send failed",FATAL);

#if 0
OLD BLOCKING METHOD
printf("%d IN sentmsg buffer:%x len:%d To:%d type:%d mid:%d\n",
      MyNodeNumber,buffer,len,ToWhom,type,SendMsgId);  */
  rc = mpc_bsend(buffer, len, ToWhom, type); /*, &SendMsgId); */
#endif

}

/* **********************************************************************/
void pcxx_MainBegin(int *argc, char *argv[])
{
  int rc;

  pcxx_InitParam (argc, argv);

  rc = mpc_environ(&MyTotalNodes,&MyNodeNumber);
  if (rc < EUI_SUCCEED)  Message("mpc_environ failed",FATAL);
 
  pcxx_InitClocks ();
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  printf("%d: pC++ initialization beginning\n",MyNodeNumber); 
  fflush(stdout);

  SendMsgId = -1; /* no messages pending */
  RecvMsgId = -1;
  pcxx_MsgBlockSize = 128; /* size in INT of max msg, (512 k) */

  mpc_sync(ALLGRP);

  /* Post a receive for the data that will eventually arrive */ 
  MyRecv();

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  barrierFlag = 0;
  ResetCollectionTraversalLevel ();
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_PROC_MAIN_CALL, 0);
}

/*******************************************************************/
void pcxx_MainEnd()
{
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_PROC_MAIN_RETURN, 0);
  pcxx_EvClose();
}

/*********************************************************************/
void pcxx_Poll()
{
  int nbytes, type, from, i;

  PROFILE_ENTER(PROFILE_Poll);

  /* Has the pending recv been filled? */
  if (VERBOSE>3) printf("%d In pcxx_Poll(), checking on status (%d) \n",
			MyNodeNumber,RecvMsgId);
  
  nbytes = mpc_status(RecvMsgId);
  
  if (nbytes == PENDING) {
    PROFILE_EXIT(PROFILE_Poll);
    return;
  }
  
  if (nbytes == INACTIVE) {
/*    printf("%d mpc_status: inactive message! mid:%d\n"
	   ,MyNodeNumber,RecvMsgId);*/
    /* Message("pcxx_Poll() aborted",FATAL); */
    PROFILE_EXIT(PROFILE_Poll);
    return;
  }
  
  /* OK! we have a message from someplace, now extract that info */
  if (nbytes > MAXBUFSIZE) {
    printf("%d received bogus buffer size %d\n",MyNodeNumber,nbytes);
    Message("Error: mpc_status returned bogus buffer size",FATAL);
  }
  
  type = MsgType;
  from = MsgFromWhom;
  
  switch(type){
    case BroadcastMsg:
      /* VERY INEFFICIENT!  Copy message from ReceiveBuffer to BlockDataBuff */
      for (i=0; i < nbytes; i++)  
	BlockDataBuff[i] = ReceiveBuffer[i];
      broadcast_flag = 1;
      break;
    case ReduceMsg:
      if(nbytes > MAXBUFSIZE) 
	Message("Reduction message too big for buffer",FATAL);
      reduction_flag++;
      PROFILE_EXIT(PROFILE_Poll);
      return;
    case HereIsMsg:
      here_is_data_handler(type, from);
      break;
    case WhereIsMsg:
      where_is_data_handler(type, from);
      break;
    case ReadBlockMsg:
      read_block_handler(type, from);
      break;
    case EndOfBlockMsg:
      /* VERY INEFFICIENT!  Copy message from ReceiveBuffer to BlockDataBuff */
      for (i=0; i < nbytes; i++)  
	BlockDataBuff[i] = ReceiveBuffer[i];
      end_of_block_handler(type, from);
      break;
    case ReadFArrayMsg:
      read_farray_handler(type, from);
      break;
    case ReadFArrayElemMsg:
      read_farray_elem_handler(type, from);
      break;
    case BarrierFanInMsg:
      BarrierFanIn(from);
      break;
    case BarrierFanOutMsg:
      /* There is nothing to "consume", just set flag, I'm done... */
      barrierFlag = DONE;
      break;
    default:
      fprintf(stderr,"Flushing message %d on node %d\n",type, 
              pcxx_NodeNumber());
    }

    /* Assume that the receive buffer has been consumed, and it is safe to
       post another receive for messages that may or maynot later arrive */

  MyRecv();
  PROFILE_EXIT(PROFILE_Poll);
  return;
}
