/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*n
/* bbnlib : BBN run-time support library			*/
/*==============================================================*/

#include <heap.h>

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

#define BCASTMAX 1024

extern int MyTotalNodes;
extern int MyNodeNumber;
extern int _TSIZE;
extern int _TDESC;
extern int _TINDEX;
extern int _CMPTDESC;

extern char *heapmalloc();
extern char *shmalloc();
extern void pcxx_Barrier();

shared void **BBN_Elem_Alloc_Vector;
shared char BroadcastBuff[BCASTMAX];  /* a place to store broacast values */

int BBN_Switch = 1;

void BBN_AllocElemVec ()
{
  BBN_Elem_Alloc_Vector = (void **) shmalloc (MyTotalNodes * sizeof(void *));
}

int SMRTSNewProc(num)
int num;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);
  BBN_Switch = 1;
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
}

void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  BBN_Switch = 1 - BBN_Switch;
  sbarrier_half(_TDESC, _TINDEX, _TSIZE, MyNodeNumber, BBN_Switch);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*
 @ pcxx_BrodcastBytes : sends messages from 0 to all other nodes
 */

void pcxx_BroadcastBytes(source, length, buffer)
int source;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  if ( length > BCASTMAX ) Message("Broadcast Too BIG!! ",FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (source == 0) for (i=0; i<length; i++) c[i] = p[i];
  pcxx_Barrier();
  if (source != 0) for (i=0; i<length; i++) p[i] = c[i];
}

void SMRTSAllocElemSpace (proc, size)
  int proc;
  int size;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_ALLOC_ELEM_ENTER, proc);
  BBN_Elem_Alloc_Vector[proc]
                 = (void *) heapmalloc (HEAP_WRITE_THRU, HEAP_LOCAL_NODE, size);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_ALLOC_ELEM_EXIT, BBN_Elem_Alloc_Vector[proc]);
}

void *SMRTSGetElemSpacePtr (proc)
  int proc;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_GET_ELEM_PTR, proc);
  return (BBN_Elem_Alloc_Vector[proc]);
}

void SMRTSDealloc(ptr)
void *ptr;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  heapfree(ptr);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}
